import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets
import numpy as np
from exporters import ImageExporter
from .default_parameters import feature_spike_opacity


def get_channel_color(cell_number):

    """This method will match the cell number with the color it should be RGB in Tint.
    These cells are numbered from 1-30 (there is technically a zeroth cell, but that isn't plotted"""

    if cell_number is None:
        raise ValueError("The cell_number cannot be none!")

    spike_colors = [(1, 8, 184), (93, 249, 75), (234, 8, 9),
                    (229, 22, 239), (80, 205, 243), (27, 164, 0),
                    (251, 188, 56), (27, 143, 167), (127, 41, 116),
                    (191, 148, 23), (185, 9, 17), (231, 223, 67),
                    (144, 132, 145), (34, 236, 228), (217, 20, 145),
                    (172, 64, 80), (176, 106, 138), (199, 194, 167),
                    (216, 204, 105), (160, 204, 61), (187, 81, 88),
                    (45, 216, 122), (242, 136, 25), (50, 164, 161),
                    (249, 67, 16), (252, 232, 147), (114, 156, 238),
                    (241, 212, 179), (129, 62, 162), (235, 133, 126)]

    while cell_number > len(spike_colors)-1:
        cell_number = cell_number - len(spike_colors)

    return spike_colors[int(cell_number)-1]


class CustomViewBox(pg.ViewBox):
    """
    Subclass of ViewBox
    """

    def __init__(self, window, item, parent=None):
        """
        Constructor of the CustomViewBox
        """
        super(CustomViewBox, self).__init__(parent)
        # self.plot = plot
        self.window = window
        self.item = item
        self.menu = None  # Override pyqtgraph ViewBoxMenu
        self.menu = self.getMenu()  # Create the menu

    def raiseContextMenu(self, ev):
        """
        Raise the context menu
        """
        if not self.menuEnabled():
            return
        menu = self.getMenu()
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))

    def getMenu(self):
        """
        Create the menu
        """
        if self.menu is None:
            self.menu = QtWidgets.QMenu()
            self.save_plot = QtWidgets.QAction("Save Figure", self.menu)
            self.save_plot.triggered.connect(self.export)
            self.menu.addAction(self.save_plot)
        return self.menu

    def export(self):
        # choose filename to save as
        save_filename, save_ext = QtWidgets.QFileDialog.getSaveFileName(QtGui.QWidget(), 'Save Scores', '',
                                                                        'PNG (*.png);;JPG (*.jpg);;TIF (*.tif);;GIF (*.gif)')

        if save_filename == '':
            return

        # create an exporter instance, as an argument give it
        # the item you wish to export

        if 'GraphicsWindow' in str(self.item):
            # get the main plot which occurs at row=1, and column=0
            plotitem = self.item.getItem(0, 0)
            # turn off the infinite line marking where the cursor is
            # self.window.mouse_vLine.hide()

            exporter = ImageExporter(plotitem)

            # set export parameters if needed
            # exporter.parameters()['width'] = 100  # (note this also affects height parameter)

            # save to file
            exporter.export(save_filename)

            # self.window.mouse_vLine.show()

        elif 'PltWidget' in str(self.item):
            plotitem = self.item.getPlotItem()

            exporter = ImageExporter(plotitem)

            # set export parameters if needed
            # exporter.parameters()['width'] = 100  # (note this also affects height parameter)

            # save to file
            exporter.export(save_filename)


class PltWidget(pg.PlotWidget):
    """
    Subclass of PlotWidget created so that we can have a custom viewBox with our own menu on right click
    """
    def __init__(self, window, parent=None):
        """
        Constructor of the widget
        """
        super(PltWidget, self).__init__(parent, viewBox=CustomViewBox(window, self))


class MultiLine(pg.QtGui.QGraphicsPathItem):
    def __init__(self, x, y, *args, **kwargs):
        """x and y are 2D arrays of shape (Nplots, Nsamples)"""
        self._boundingRect = None

        connect = np.ones(x.shape, dtype=bool)
        connect[:, -1] = 0  # don't draw the segment between each trace
        self.path = pg.arrayToQPath(x.flatten(), y.flatten(), connect.flatten())

        pg.QtGui.QGraphicsPathItem.__init__(self, self.path)

        self.check_function_call = 0

        pen_kwargs = {}
        for kwarg in kwargs:
            if 'pen_' in kwarg:
                kwargNew = kwarg.split('pen_')[1]
                pen_kwargs[kwargNew] = kwargs[kwarg]

        if len(pen_kwargs) > 0:
            self.setPen(pg.mkPen(**pen_kwargs))
        else:
            self.setPen(pg.mkPen('w'))

    def shape(self):  # override because QGraphicsPathItem.shape is too expensive.
        return pg.QtGui.QGraphicsItem.shape(self)

    def boundingRect(self):
        if self._boundingRect is None:
            self._boundingRect = self.path.boundingRect()
        return self._boundingRect


def get_spike_colors(self):
    """
    This function will get the colors of the spike so we can plot them using the same colors that Tint uses.

    :param self: the main window object
    :return: None
    """
    n_spikes = self.tetrode_data.shape[1]

    self.spike_colors = np.ones((n_spikes, 4))
    self.spike_colors[:, -1] = feature_spike_opacity

    unique_cells = np.unique(self.cut_data)
    # the 0'th cell is the dummy cell for Tint so we will remove that
    unique_cells = unique_cells[unique_cells != 0]

    self.spike_colors[np.where(self.cut_data == 0)[0], :] = np.zeros(4)

    for cell in unique_cells:
        cell_color = get_channel_color(cell)
        cell_bool = np.where(self.cut_data == cell)[0]
        self.spike_colors[cell_bool, :-1] = np.asarray(cell_color)/255





import os
import numpy as np


def write_cut(cut_filename, cut, basename=None):
    if basename is None:
        basename = os.path.basename(os.path.splitext(cut_filename)[0])

    unique_cells = np.unique(cut)

    if 0 not in unique_cells:
        # if it happens that there is no zero cell, add it anyways
        unique_cells = np.insert(unique_cells, 0, 0)  # object, index, value to insert

    n_clusters = len(np.unique(cut))
    n_spikes = len(cut)

    write_list = []  # the list of values to write

    tab = '    '  # the spaces didn't line up with my tab so I just created a string with enough spaces
    empty_space = '               '  # some of the empty spaces don't line up to x tabs

    # we add 1 to n_clusters because zero is the garbage cell that no one uses
    write_list.append('n_clusters: %d\n' % (n_clusters))
    write_list.append('n_channels: 4\n')
    write_list.append('n_params: 2\n')
    write_list.append('times_used_in_Vt:%s' % ((tab + '0') * 4 + '\n'))

    zero_string = (tab + '0') * 8 + '\n'

    for cell_i in np.arange(n_clusters):
        write_list.append(' cluster: %d center:%s' % (cell_i, zero_string))
        write_list.append('%smin:%s' % (empty_space, zero_string))
        write_list.append('%smax:%s' % (empty_space, zero_string))
    write_list.append('\nExact_cut_for: %s spikes: %d\n' % (basename, n_spikes))

    # now the cut file lists 25 values per row
    n_rows = int(np.floor(n_spikes / 25))  # number of full rows

    remaining = int(n_spikes - n_rows * 25)
    cut_string = ('%3u' * 25 + '\n') * n_rows + '%3u' * remaining

    write_list.append(cut_string % (tuple(cut)))

    with open(cut_filename, 'w') as f:
        f.writelines(write_list)


def write_clu(clu_filename, data):
    # the .clu files and the .cut files are different since the .clu files are the .cut files (with no manual sorting)
    # without the headers, and the values go from 1 -> N instead of 0 -> N, (1-based numbering instead of 0-based). Thus
    # we add 1 to the .cut data to get the .clu data

    data = np.asarray(data).astype(int)  # ensuring that the data is the integer data-type

    data += 1  # making the data 1-based instead of 0-based

    # calculating the number of clusters
    n_clust = len(np.unique(data))

    # ensuring that the cluster number is the 1st value
    data = np.concatenate(([n_clust], data))

    # saving the data as a column (delimter='\n') and integer format.
    np.savetxt(clu_filename, data, fmt='%d', delimiter='\n')



from .feature_functions import CreateFeatures
from .default_parameters import openGL, gridLines, feature_spike_size
from .plot_utils import get_spike_colors
import pyqtgraph.opengl as gl
import numpy as np

feature_name_map = {
    # 'PC1': 'WavePCX!1',
    # 'PC2': 'WavePCX!2',
    # 'PC3': 'WavePCX!3',
    # 'PC4': 'WavePCX!4',
    'PC1': 'WavePCX_scikit!1',
    'PC2': 'WavePCX_scikit!2',
    'PC3': 'WavePCX_scikit!3',
    'PC4': 'WavePCX_scikit!4',
    'Energy': 'Energy',
    'Amplitude': 'Amplitude',
    'Peak': 'Peak',
    'Peak Time': 'PeakTime',
    'Trough': 'Trough',
    'Trough Time': 'TroughTime',
}


def load_features(self):
    """
    This functionn will calculate the features that the user has chosen in the Main Window and save it in the
    self.feature_data dictionary

    :param self:
    :return:
    """

    for feature in [self.x_axis_cb.currentText(), self.y_axis_cb.currentText(), self.z_axis_cb.currentText()]:

        if feature == 'None':
            continue

        feature_function = feature_name_map[feature]

        if self.feature_data is None:
            self.feature_data = {}

        if feature not in self.feature_data.keys():
            self.feature_data[feature] = CreateFeatures(self.tetrode_data, featuresToCalculate=[feature_function])


def plot_features(self):
    """
    This function will take the feature list from the Main Window and plot the data onto the self.feature_win
    window.

    :param self: the main window object
    :return: None
    """

    feature_list = [self.x_axis_cb.currentText(), self.y_axis_cb.currentText(), self.z_axis_cb.currentText()]

    feature_list = [feature for feature in feature_list if feature != 'None']

    for channel in np.arange(self.n_channels):
        Zaxis = False
        data = None
        if len(feature_list) > 2:
            # 3d graph
            feature_x = feature_list[0]
            X = self.feature_data[feature_x][:, channel]

            feature_y = feature_list[1]
            Y = self.feature_data[feature_y][:, channel]

            feature_z = feature_list[2]
            Z = self.feature_data[feature_z][:, channel]

            Zaxis = True

            data = np.vstack((X, Y, Z))
        elif len(feature_list) <= 1:
            return
        else:
            # 2d graph
            feature_x = feature_list[0]
            X = self.feature_data[feature_x][:, channel]
            feature_y = feature_list[1]
            Y = self.feature_data[feature_y][:, channel]
            feature_z = ''
            data = np.vstack((X, Y))

        if openGL:
            # --------------- this is if we want to use the open gl 3d plots ---------------------------- #

            # Add the axis lines to the plot to provide users with orientation
            if gridLines:

                startx = np.amin(X)
                starty = np.amin(Y)

                if Zaxis:
                    startz = np.amin(Z)
                else:
                    startz = 0

                if self.xline is None:
                    self.xline = gl.GLLinePlotItem(pos=np.array([[startx, starty, startz], [np.amax(X), starty, startz]]),
                                                   color=(1, 0, 0, 1), width=2, antialias=True)
                    self.glViewWidget.addItem(self.xline)
                else:
                    self.xline.setData(pos=np.array([[startx, starty, startz], [np.amax(X), starty, startz]]),
                                       color=(1, 0, 0, 1), width=2, antialias=True)

                if self.yline is None:
                    self.yline = gl.GLLinePlotItem(pos=np.array([[startx, starty, startz], [startx, np.amax(Y), startz]]),
                                                   color=(0, 1, 0, 1), width=2, antialias=True)
                    self.glViewWidget.addItem(self.yline)
                else:
                    self.yline.setData(pos=np.array([[startx, starty, startz], [startx, np.amax(Y), startz]]),
                                       color=(0, 1, 0, 1), width=2, antialias=True)

                if Zaxis:
                    if self.zline is None:
                        self.zline = gl.GLLinePlotItem(pos=np.array([[startx, starty, startz], [startx, starty, np.amax(Z)]]),
                                                       color=(0, 0, 1, 1), width=2, antialias=True)
                        self.glViewWidget.addItem(self.zline)
                    else:
                        self.zline.setData(pos=np.array([[startx, starty, startz], [startx, starty, np.amax(Z)]]),
                                           color=(0, 0, 1.0, 0.5), width=2, antialias=True)
                        self.zline.show()
                else:
                    if self.zline is not None:
                        self.zline.hide()

            if self.spike_colors is None:
                get_spike_colors(self)

            if self.scatterItem is not None:
                self.scatterItem.setData(pos=data.T, color=self.spike_colors,
                                         size=feature_spike_size)
            else:
                self.scatterItem = gl.GLScatterPlotItem(pos=data.T, color=self.spike_colors,
                                                        size=feature_spike_size)
                self.glViewWidget.addItem(self.scatterItem)



import os
import time
import numpy as np
from .default_parameters import channel_range, max_num_actions
from .gui_utils import validate_session
from .Tint_Matlab import getspikes, read_cut, read_clu
from .plot_utils import CustomViewBox, get_channel_color, MultiLine
from .feature_plot import load_features, plot_features
from .waveform_cut_functions import findSpikeSubsample, get_index_from_cell, \
    get_index_from_old_cell, get_cell_from_index, setPlotTitle, get_channel_from_y, validateMoveValue, \
    find_spikes_crossed, get_max_spikes, get_next_action, get_old_index_from_position, get_grid_dimensions, clear_unit
import pyqtgraph as pg
from functools import partial
from PyQt5 import QtGui
from pyqtgraph.Qt import QtCore


def replot_unit(self, index, cell=None):
    """
    This function will re-plot the cell. This is likely called when you have indicated that you wanted to remove the
    spikes from the cell to another value.

    :param self:
    :param index:
    :param cell:
    :return:
    """

    if index is None:
        return

    # get the cell for the color
    if cell is None:
        cell = get_cell_from_index(self, index)

    # skip the 0'th cell, that is the dummy cell in Tint
    if cell == 0:
        return

    setTitle = False
    for channel in np.arange(self.n_channels):
        if index in self.plot_lines.keys():
            self.unit_plots[index][0].removeItem(self.plot_lines[index][channel])
            self.unit_plots[index][0].removeItem(self.avg_plot_lines[index][channel])

        plot_data = self.unit_data[index][channel]

        current_n = plot_data.shape[0]

        plot_data = plot_data[self.cell_subsample_i[cell][channel], :]

        self.plot_lines[index][channel] = MultiLine(
            np.tile(np.arange(self.samples_per_spike), (plot_data.shape[0], 1)),
            plot_data, pen_color=get_channel_color(cell))

        self.avg_plot_lines[index][channel] = MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                               np.mean(self.unit_data[index][channel], axis=0).reshape((1, -1)),
                                               pen_color='w', pen_width=2)

        self.unit_plots[index][0].addItem(self.plot_lines[index][channel])
        self.unit_plots[index][0].addItem(self.avg_plot_lines[index][channel])

        if not setTitle:
            if cell in self.original_cell_count.keys():
                setPlotTitle(self.unit_plots[index][0], cell, original_cell_count=self.original_cell_count[cell],
                             current_cell_count=current_n)
            else:
                setPlotTitle(self.unit_plots[index][0], cell, current_cell_count=current_n)
            setTitle = True

    # check if there is a popup window for this cell
    if cell in self.PopUpCutWindow.keys():
        # check if the popup window for this cell is active
        if self.PopUpCutWindow[cell].PopUpActive:
            # if self.PopUpCutWindow[cell].cell == cell:
            self.PopUpCutWindow[cell].plot(index, cell)  # re-plot on the popup window


def reconfigure_units(self, unique_cells):
    """
    This function is called when the cell that you want to plot has not been created yet. Thus we might need to
    reconfigure the subplots (add a row or column).

    :param self:
    :param unique_cells:
    :return:
    """

    unique_cells = sorted(unique_cells)

    # number of cells to plot (this includes the newly added cell)
    n_cells = len(unique_cells)

    # the number of rows and columns that we want to have plotted
    rows, cols = get_grid_dimensions(n_cells, method='nper', n=3)

    # initialize the current row and column count
    row = 0
    col = 0

    # we will create a dictionary to contain the old spike plots
    old_plots = {}
    for index, value in self.plot_lines.items():
        old_plots[index] = value
    self.plot_lines = {}

    # we will create a dictionary to contain the old average plots
    old_avgs = {}
    for index, value in self.avg_plot_lines.items():
        old_avgs[index] = value
    self.avg_plot_lines = {}

    # we will create a dictionary to contain the old unit data
    old_unit_data = {}
    for index, value in self.unit_data.items():
        old_unit_data[index] = value
    self.unit_data = {}
    self.unit_data[-1] = old_unit_data[-1]  # preserving the 0'th cell

    # we will create a dictionary to contain the old average plots
    old_unit_plots = {}
    old_plotted_units = []
    for index, value in self.unit_plots.items():
        old_unit_plots[index] = value
        if value[1] != 0:
            old_plotted_units.append(value[1])  # append the cell value
    self.unit_plots = {}
    self.unit_plots[-1] = old_unit_plots[-1]  # preserving the 0'th cell

    # get the old positions of the plots (row, col)
    self.old_positions = {}
    for index, value in self.unit_positions.items():
        self.old_positions[index] = value
    self.unit_positions = {}

    # get the old ROI objects
    old_unit_drag_lines = {}
    for index, value in self.unit_drag_lines.items():
        old_unit_drag_lines[index] = value
        value.hide()
    self.unit_drag_lines = {}

    # get the old viewbox objects
    old_vb = {}
    for index, value in self.vb.items():
        old_vb[index] = value
    self.vb = {}

    max_spike_plots = get_max_spikes(self)
    if self.max_spike_plots is None:
        if max_spike_plots is None:
            return
        else:
            self.max_spike_plots = max_spike_plots

    for index in np.arange(rows*cols):

        if index < len(unique_cells):
            # then we should re-plot these subplot indices

            cell = unique_cells[index]

            # first we need to determine if we need to add a plot or not
            if (row, col) not in self.old_positions.keys():
                '''
                The plot does not exist in this position before, so we will need to add a subplot in this position
                configure the plot appropriately. As well as add another polyline ROI, and define the drag and mouse
                clicks
                '''

                # add the plot
                self.unit_positions[(row, col)] = index
                self.unit_plots[index] = [self.unit_win.addPlot(row=row, col=col,
                                                                viewBox=CustomViewBox(self, self.unit_win)), cell]
                # add the plot
                self.vb[index] = self.unit_plots[index][0].vb

                # configure the plot
                self.unit_plots[index][0].setXRange(0, self.samples_per_spike, padding=0)  # set the x-range
                self.unit_plots[index][0].setYRange(0, -self.n_channels * channel_range, padding=0)
                self.unit_plots[index][0].hideAxis('left')  # remove the y-axis
                self.unit_plots[index][0].hideAxis('bottom')  # remove the x axis
                self.unit_plots[index][0].hideButtons()  # hide the auto-resize button
                self.unit_plots[index][0].setMouseEnabled(x=False, y=False)  # disables the mouse interactions
                self.unit_plots[index][0].enableAutoRange(False, False)

                # add the ROI
                self.unit_drag_lines[index] = pg.PolyLineROI([[0, 0], [30, 30]])
                self.unit_drag_lines[index].hide()
                self.unit_plots[index][0].addItem(self.unit_drag_lines[index])

                # add the drag and mouse click events to the viewbox
                self.vb[index].mouseDragEvent = partial(drag, self, index)  # overriding the drag event
                self.vb[index].mouseClickEvent = partial(mouse_click_event, self, index)

            else:
                '''
                The row/column does exist. we won't need to create a new subplot, but we will need to move
                everything over and delete the old contents.
                '''
                # we will need to find the old index value so we can make sure we can access that plot
                old_plot_index = get_old_index_from_position(self, (row, col))  # get the index

                # defining the new plot use
                self.unit_plots[index] = [old_unit_plots[old_plot_index][0], cell]
                self.unit_positions[(row, col)] = index

                self.vb[index] = self.unit_plots[index][0].vb
                self.vb[index].mouseDragEvent = partial(drag, self, index)  # overriding the drag event
                self.vb[index].mouseClickEvent = partial(mouse_click_event, self, index)

                self.unit_drag_lines[index] = old_unit_drag_lines[old_plot_index]

            if cell in old_plotted_units:
                # figure out if we have plotted this cell before, if so, we can move that data instead of re-plotting

                old_index = get_index_from_old_cell(cell, old_unit_plots)

                current_n = len(self.cell_indices[cell])
                if cell in self.original_cell_count.keys():
                    setPlotTitle(self.unit_plots[index][0], cell,
                                 original_cell_count=self.original_cell_count[cell],
                                 current_cell_count=current_n)
                else:
                    setPlotTitle(self.unit_plots[index][0], cell, current_cell_count=current_n)

                # we are defining the new spike plots and avg plots given the old index
                self.plot_lines[index] = old_plots[old_index]
                self.avg_plot_lines[index] = old_avgs[old_index]

                # moving the line plots
                for channel in self.plot_lines[index].keys():
                    old_unit_plots[old_index][0].removeItem(old_plots[old_index][channel])
                    self.unit_plots[index][0].addItem(self.plot_lines[index][channel])

                # moving the avg plots
                for channel in self.avg_plot_lines[index].keys():
                    old_unit_plots[old_index][0].removeItem(old_avgs[old_index][channel])
                    self.unit_plots[index][0].addItem(self.avg_plot_lines[index][channel])

                # moving the unit data
                self.unit_data[index] = old_unit_data[old_index]

            else:
                # This cell has not been plotted before, so we will need to create the MultiLines

                setTitle = False
                for channel in np.arange(self.n_channels):

                    self.unit_positions[(row, col)] = index

                    # shifting the data so that the next channel resides below the previous
                    # also making the 1st channel start at a y value of 0

                    cell_bool = self.cell_indices[cell]
                    cell_data = self.tetrode_data[:, cell_bool, :]

                    plot_data = cell_data[channel]

                    channel_max = 127
                    plot_data = plot_data - channel * channel_range - channel_max

                    current_n = plot_data.shape[0]
                    if not setTitle:
                        if cell in self.original_cell_count.keys():
                            setPlotTitle(self.unit_plots[index][0], cell,
                                         original_cell_count=self.original_cell_count[cell],
                                         current_cell_count=current_n)
                        else:
                            setPlotTitle(self.unit_plots[index][0], cell, current_cell_count=current_n)
                        setTitle = True

                    cell_data_sub_channel, subsample_i = findSpikeSubsample(plot_data, self.max_spike_plots)

                    if cell not in self.cell_subsample_i.keys():
                        self.cell_subsample_i[cell] = {channel: subsample_i}
                    else:
                        self.cell_subsample_i[cell][channel] = subsample_i

                    plot_data_avg = np.mean(plot_data, axis=0).reshape((1, -1))

                    if index not in self.unit_data.keys():
                        self.unit_data[index] = {channel: plot_data}
                    else:
                        self.unit_data[index][channel] = plot_data

                    if index not in self.plot_lines.keys():

                        self.plot_lines[index] = {
                            channel: MultiLine(
                                np.tile(np.arange(self.samples_per_spike), (cell_data_sub_channel.shape[0], 1)),
                                cell_data_sub_channel, pen_color=get_channel_color(cell))}
                    else:
                        self.plot_lines[index][channel] = MultiLine(
                            np.tile(np.arange(self.samples_per_spike), (cell_data_sub_channel.shape[0], 1)),
                            cell_data_sub_channel, pen_color=get_channel_color(cell))

                    if index not in self.avg_plot_lines.keys():
                        self.avg_plot_lines[index] = {
                            channel: MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                               plot_data_avg, pen_color='w', pen_width=2)}
                    else:
                        self.avg_plot_lines[index][channel] = MultiLine(
                            np.arange(self.samples_per_spike).reshape((1, -1)),
                            plot_data_avg, pen_color='w', pen_width=2)

                    self.unit_plots[index][0].addItem(self.plot_lines[index][channel])
                    self.unit_plots[index][0].addItem(self.avg_plot_lines[index][channel])
        else:
            # i made this to remove the items from the plots that we don't need anymore, however I end up just
            # removing the subplot itself so this is no longer necessary
            '''
            # then we should check if the plot has any data on it, and remove it since we no longer need it
            if (row, col) in self.old_positions.keys():
                # then we had previously plotted stuff here
                old_plot_index = get_old_index_from_position(self, (row, col))  # get the index

                for channel in old_plots[old_plot_index].keys():
                    old_unit_plots[old_plot_index][0].removeItem(old_plots[old_plot_index][channel])

                # removing the avg plots
                for channel in old_avgs[old_plot_index].keys():
                    old_unit_plots[old_plot_index][0].removeItem(old_avgs[old_plot_index][channel])

            else:
                # there is nothing plotted, so don't worry about it
                pass
            '''
            pass

        col += 1
        if col >= cols:
            col = 0
            row += 1

    # removing any sub_plots that are no longer necessary
    for position in self.old_positions.keys():
        if position not in self.unit_positions.keys():
            old_plot_index = get_old_index_from_position(self, position)  # get the i
            plot_item = old_unit_plots[old_plot_index][0]
            self.unit_win.removeItem(plot_item)

    # removing any old plots that are no longer necessary
    for index, value in old_unit_plots.items():
        cell = value[1]
        if cell not in self.cell_indices.keys():
            for channel in old_plots[index].keys():
                old_unit_plots[index][0].removeItem(old_plots[index][channel])
                old_unit_plots[index][0].removeItem(old_avgs[index][channel])

    self.old_positons = {}
    self.unit_rows = rows
    self.unit_cols = cols


def plot_units(self):
    """
    This function will go and plot all the cells (besides the 0'th as that is the dummy cell).
    It will create the subplots in the self.unit_win plot object. Find the data and indices for each of
    the cells and then plot them.

    :param self:
    :return:
    """
    unique_cells = np.unique(self.cut_data)
    # the 0'th cell is the dummy cell for Tint so we will remove that

    unique_cells = unique_cells[unique_cells != 0]

    cell_bool = np.where(self.cut_data == 0)[0]
    cell_data = self.tetrode_data[:, cell_bool, :]
    self.cell_indices[0] = np.where(self.cut_data == 0)[0]

    for channel, channel_data in enumerate(cell_data):
        self.unit_plots[-1] = [None, 0]
        if -1 not in self.unit_data.keys():
            self.unit_data[-1] = {channel: channel_data}
        else:
            self.unit_data[-1][channel] = channel_data

    n_cells = len(unique_cells)

    rows, cols = get_grid_dimensions(n_cells, method='nper', n=3)

    if self.max_spike_plots is None:
        max_spike_plots = get_max_spikes(self)
        if max_spike_plots is None:
            return
        else:
            self.max_spike_plots = max_spike_plots

    # add the dummy 0'th cell to the cell_indices for the undo functionality to work properly

    row = 0
    col = 0
    for index in np.arange(len(unique_cells)):
        self.unit_positions[(row, col)] = index

        cell = unique_cells[index]

        self.unit_plots[index] = [self.unit_win.addPlot(row=row, col=col,
                                                        viewBox=CustomViewBox(self, self.unit_win)), cell]

        self.vb[index] = self.unit_plots[index][0].vb

        self.unit_plots[index][0].setXRange(0, self.samples_per_spike, padding=0)  # set the x-range
        self.unit_plots[index][0].setYRange(0, -self.n_channels * channel_range, padding=0)
        self.unit_plots[index][0].hideAxis('left')  # remove the y-axis
        self.unit_plots[index][0].hideAxis('bottom')  # remove the x axis
        self.unit_plots[index][0].hideButtons()  # hide the auto-resize button
        self.unit_plots[index][0].setMouseEnabled(x=False, y=False)  # disables the mouse interactions
        self.unit_plots[index][0].enableAutoRange(False, False)

        self.unit_drag_lines[index] = pg.PolyLineROI([[0, 0], [30, 30]])
        self.unit_drag_lines[index].hide()
        self.unit_plots[index][0].addItem(self.unit_drag_lines[index])

        self.vb[index].mouseDragEvent = partial(drag, self, index)  # overriding the drag event
        self.vb[index].mouseClickEvent = partial(mouse_click_event, self, index)

        cell_bool = np.where(self.cut_data == cell)[0]
        cell_data = self.tetrode_data[:, cell_bool, :]

        self.original_cell_count[cell] = cell_data.shape[1]

        if self.samples_per_spike is None:
            self.samples_per_spike = cell_data.shape[2]

        if self.n_channels is None:
            self.n_channels = cell_data.shape[0]

        self.cell_indices[cell] = cell_bool

        # cell_data_sub = findSpikeSubsample(cell_data, self.max_spike_plots)

        title_set = False
        for channel in np.arange(self.n_channels):
            # we will keep the data and the indices

            # shifting the data so that the next channel resides below the previous
            # also making the 1st channel start at a y value of 0
            plot_data = cell_data[channel]
            # channel_max = np.amax(plot_data)
            channel_max = 127
            plot_data = plot_data - channel*channel_range - channel_max
            if not title_set:
                setPlotTitle(self.unit_plots[index][0], cell, original_cell_count=self.original_cell_count[cell],
                             current_cell_count=plot_data.shape[0])
                title_set = True

            cell_data_sub_channel, subsample_i = findSpikeSubsample(plot_data, self.max_spike_plots)

            if cell not in self.cell_subsample_i.keys():
                self.cell_subsample_i[cell] = {channel: subsample_i}
            else:
                self.cell_subsample_i[cell][channel] = subsample_i

            plot_data_avg = np.mean(plot_data, axis=0).reshape((1, -1))

            if index not in self.unit_data.keys():
                self.unit_data[index] = {channel: plot_data}
            else:
                self.unit_data[index][channel] = plot_data

            if index not in self.plot_lines.keys():

                self.plot_lines[index] = {channel: MultiLine(np.tile(np.arange(self.samples_per_spike), (cell_data_sub_channel.shape[0], 1)),
                                                             cell_data_sub_channel, pen_color=get_channel_color(cell))}
            else:
                self.plot_lines[index][channel] = MultiLine(np.tile(np.arange(self.samples_per_spike), (cell_data_sub_channel.shape[0], 1)),
                                                            cell_data_sub_channel, pen_color=get_channel_color(cell))

            if index not in self.avg_plot_lines.keys():
                self.avg_plot_lines[index] = {channel: MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                                                 plot_data_avg, pen_color='w', pen_width=2)}
            else:
                self.avg_plot_lines[index][channel] = MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                                                plot_data_avg, pen_color='w', pen_width=2)

            self.unit_plots[index][0].addItem(self.plot_lines[index][channel])
            self.unit_plots[index][0].addItem(self.avg_plot_lines[index][channel])

        col += 1
        if col >= cols:
            col = 0
            row += 1

    self.unit_rows = rows
    self.unit_cols = cols


def plot_session(self):

    self.reset_parameters()

    session_filename = self.filename.text()

    session_valid, error_raised = validate_session(self, session_filename)

    if session_valid:
        tetrode = self.tetrode_cb.currentText()

        session_filepath = os.path.splitext(self.filename.text())[0]

        tetrode_filename = '%s.%s' % (session_filepath, tetrode)

        if os.path.exists(tetrode_filename):

            if self.tetrode_data is None:

                ts, ch1, ch2, ch3, ch4, spikeparams = getspikes(tetrode_filename)
                self.tetrode_data = np.vstack((ch1, ch2, ch3, ch4)).reshape((4, -1, ch1.shape[1]))
                if self.spike_times is None:
                    self.spike_times = ts

                self.n_channels = self.tetrode_data.shape[0]
                self.samples_per_spike = self.tetrode_data.shape[2]

            if self.cut_data is None:
                filename = self.cut_filename.text()  # the data filename (could be .cut or .clu file)
                if '.clu.' in filename:
                    cut_data = read_clu(filename)
                else:
                    cut_data = read_cut(filename)

                self.cut_data = cut_data
                self.cut_data_original = self.cut_data.copy()  # keep a copy of the original to revert if we want.

            load_features(self)

            plot_features(self)

            plot_units(self)

        else:
            # filename does not exist
            self.choice = None
            self.LogError.signal.emit('TetrodeExistError!%s' % tetrode_filename)

            while self.choice is None:
                time.sleep(0.1)

        return


def drag(self, index, ev=None):
    """
    This function will allow us to modify the line ROI's that we created for each Unity graph. This is so
    the users can select which spikes to cut. We have a dictionary containing each of the ROI's (self.unit_drag_lines),
    where the index (the plot index) is the key, and the value is ROI object.

    Each plot has a different corresponding index as the input to drag, so we can make sure to manipulate the correct
    ROI object using the index input.

    self: the main window object containing the attributes that we will be needing
    index: the index value of the plot (order that the plots were created essentially)
    ev: event.
    """

    # global vb, lr
    if ev.button() == QtCore.Qt.LeftButton:
        # the user is using the left
        if not self.drag_active or index != self.last_drag_index:
            for roi in self.active_ROI:
                try:
                    roi.hide()
                except RuntimeError:
                    # likely the subplot was removed
                    pass

            self.unit_drag_lines[index].show()  # showing the LineSegmentROI
            self.active_ROI = [self.unit_drag_lines[index]]
            self.drag_active = True

        # defining the start of the selected region
        points = [[self.vb[index].mapToView(ev.buttonDownPos()).x(),
                  self.vb[index].mapToView(ev.buttonDownPos()).y()],
                  [self.vb[index].mapToView(ev.pos()).x(),
                   self.vb[index].mapToView(ev.pos()).y()]]

        self.unit_drag_lines[index].setPoints(points)
        self.last_drag_index = index
        ev.accept()
    else:
        pg.ViewBox.mouseDragEvent(self.vb[index], ev)


def mouse_click_event(self, index, ev=None):
    """
    This function will override the mouse click event. If the user shift + left mouse clicks on a plot, this will
    launch the PopUpCutWindow (stored as self.PopUpCutWindow), so the user can have a better look at the data.

    If the user middle mouse click it will initiate a cutting of the cell. For this to work the user must have
    had the ROI line intersecting with the data on the plot. It will calculate the slope of the ROI line and
    then determine if this line intersects with any of the data. If the line intersects with the data, it will remove
    the data from this spike, and move it to the self.move_to_channel() attribute from the main Window. It will
    then re-plot both the cell that the spikes were removed from, and the cell tha tthe spikes were moved to.

    :param self: the main window so we can grab any attributes that we need.
    :param index: the index of the plots, essentially the order of which the plots were created
    :param ev: the event that caused this function to run
    :return: None
    """

    # determine if the user used any modifiers (shift, ctrl, etc).
    modifiers = QtGui.QApplication.keyboardModifiers()

    if ev.button() == QtCore.Qt.RightButton:
        # open menu
        pg.ViewBox.mouseClickEvent(self.vb[index], ev)

    elif ev.button() == QtCore.Qt.LeftButton:
        # we want to make sure the user shift + left clicks to initiate the pop up, check that the modifier is a
        # shift modifier.
        if modifiers == QtCore.Qt.ShiftModifier:
            # then you will launch the popup

            cell = get_cell_from_index(self, index)

            if cell not in self.PopUpCutWindow.keys():
                # this cell does not have a popup window, create one
                self.addPopup(cell)
            else:
                # reset the data for this cell
                self.PopUpCutWindow[cell].reset_data()  # clears any old data that might have been on the popup
            self.PopUpCutWindow[cell].plot(index, get_cell_from_index(self, index))  # needs index, cell as inputs
            self.PopUpCutWindow[cell].raise_()

        else:
            # hopefully drag event
            pg.ViewBox.mouseClickEvent(self.vb[index], ev)

    elif ev.button() == QtCore.Qt.MiddleButton:
        # then we will accept the changes

        if self.unit_drag_lines[index] in self.active_ROI:
            # then we have an active ROI
            # we will get the x,y positions (rounded to the nearest int) of the selected line

            cell = get_cell_from_index(self, index)

            try:
                invalid_cell_number = int(self.move_to_channel.text())
            except:
                self.choice = None
                self.LogError.signal.emit('InvalidMoveChannel')
                while self.choice is None:
                    time.sleep(0.1)
                return

            if not validateMoveValue(invalid_cell_number):
                self.choice = None
                self.LogError.signal.emit('InvalidMoveChannel')
                while self.choice is None:
                    time.sleep(0.1)
                return

            if invalid_cell_number == cell:
                self.choice = None
                self.LogError.signal.emit('SameChannelInvalid')
                while self.choice is None:
                    time.sleep(0.1)
                return

            points = np.rint(np.asarray(self.unit_drag_lines[index].getState()['points']))

            # find which channel the user started in
            channel = get_channel_from_y(points[0, 1], channel_range=channel_range, n_channels=self.n_channels)

            unit_data = self.unit_data[index][channel]

            crossed_cells = find_spikes_crossed(points, unit_data, samples_per_spike=self.samples_per_spike)

            # append the crossed lines to the invalid cell's plot
            invalid_index = get_index_from_cell(self, invalid_cell_number)

            max_spikes_changed = False
            # self.max_spike_plots will be none if the value has been changed recently
            if self.max_spike_plots is None:
                max_spike_plots = get_max_spikes(self)
                if max_spike_plots is None:
                    return
                else:
                    self.max_spike_plots = max_spike_plots

                    # update the max spikes by re-calculating the subsample index values
                    for plotted_cell in self.cell_subsample_i.keys():
                        plotted_channel_index = get_index_from_cell(self, plotted_cell)

                        if any(plotted_cell == cell_value for cell_value in [0]):
                            # skip the invalid channel (done later) and the dummy channel (not plotted)
                            continue

                        for data_chan in np.arange(self.n_channels):
                            _, subsample_i = findSpikeSubsample(self.unit_data[plotted_channel_index][data_chan],
                                                                self.max_spike_plots)

                            if plotted_cell not in self.cell_subsample_i.keys():
                                self.cell_subsample_i[plotted_cell] = {data_chan: subsample_i}
                            else:
                                self.cell_subsample_i[plotted_cell][data_chan] = subsample_i
                    max_spikes_changed = True

            reconfigure = False

            if len(crossed_cells) != 0:
                # remove these spikes from all the channels
                for data_chan in np.arange(self.n_channels):
                    if invalid_index is not None:
                        # get the invalid data
                        invalid_cell_data = self.unit_data[index][data_chan][crossed_cells, :]
                        # update the invalid_data channel with this current data
                        self.unit_data[invalid_index][data_chan] = np.vstack((self.unit_data[invalid_index][data_chan],
                                                                              invalid_cell_data))

                        # update the plotted subsample as well
                        _, subsample_i = findSpikeSubsample(self.unit_data[invalid_index][data_chan], self.max_spike_plots)
                        if invalid_cell_number not in self.cell_subsample_i.keys():
                            self.cell_subsample_i[invalid_cell_number] = {data_chan: subsample_i}
                        else:
                            self.cell_subsample_i[invalid_cell_number][data_chan] = subsample_i
                    else:
                        reconfigure = True

                    # delete the invalid data from the selected channel
                    self.unit_data[index][data_chan] = np.delete(self.unit_data[index][data_chan], crossed_cells, axis=0)

                    # recalculate subplot for the channel that the spikes were removed from
                    if len(self.unit_data[index][data_chan]) > 0:
                        _, subsample_i = findSpikeSubsample(self.unit_data[index][data_chan], self.max_spike_plots)
                        if cell not in self.cell_subsample_i.keys():
                            self.cell_subsample_i[cell] = {data_chan: subsample_i}
                        else:
                            self.cell_subsample_i[cell][data_chan] = subsample_i
                    else:
                        # there is no data left, don't need to worry about the subsampling anymore
                        if cell in self.cell_subsample_i.keys():
                            self.cell_subsample_i.pop(cell)
                            reconfigure = True

                # check if the cell still exists
                for key, value in self.unit_data[index].items():
                    if len(value) == 0:
                        if index != -1:
                            # avoid popping the dummy cell (index -1)
                            self.unit_data.pop(index)
                            reconfigure = True
                            break

                # update the bool
                cell_indices = self.cell_indices[cell]
                # append invalid cells to the new cell number
                invalid_cells = cell_indices[crossed_cells]
                self.cell_indices[cell] = np.delete(cell_indices, crossed_cells)

                # check if there are still indices for this cell, if empty we will remove
                if len(self.cell_indices[cell]) == 0:
                    clear_unit(self, cell)  # delete the cell's plots
                    if cell in self.original_cell_count.keys():
                        self.original_cell_count.pop(cell)
                    reconfigure = True

                    if cell != 0:
                        self.cell_indices.pop(cell)

                if invalid_cell_number in self.cell_indices.keys():
                    # the cell has existed already within the main window, we can just add to this plot
                    self.cell_indices[invalid_cell_number] = np.concatenate((self.cell_indices[invalid_cell_number],
                                                                             invalid_cells))
                else:
                    # this cell is not already plotted, have to add the plot and possibly reconfigure
                    self.cell_indices[invalid_cell_number] = invalid_cells

                    if invalid_cell_number != 0:
                        reconfigure = True

                # add the latest action
                if len(self.latest_actions) == 0 or max_num_actions == 1:
                    self.latest_actions = {0: {'action': 'cut', 'fromCell': cell, 'toCell': invalid_cell_number,
                                               'movedCutIndices': invalid_cells}}
                else:
                    next_action = get_next_action(self)
                    self.latest_actions[next_action] = {'action': 'cut', 'fromCell': cell, 'toCell': invalid_cell_number,
                                                        'movedCutIndices': invalid_cells}

            if not reconfigure:
                # update plots for the invalid cell and the cell you removed these spikes from
                # no need to reconfigure
                replot_unit(self, index)
                invalid_index = get_index_from_cell(self, invalid_cell_number)
                if invalid_index != -1:
                    replot_unit(self, invalid_index)
            else:
                # we will need to reconfigure the main window possibly, do so

                if cell in self.cell_indices.keys():
                    replot_unit(self, index)

                unique_cells = np.asarray(list(self.cell_indices.keys()))
                reconfigure_units(self, list(unique_cells[unique_cells != 0]))

            if max_spikes_changed:
                for plotted_cell in self.cell_indices.keys():
                    if plotted_cell == invalid_cell_number or plotted_cell == cell:
                        continue

                    plotted_cell_index = get_index_from_cell(self, plotted_cell)
                    replot_unit(self, plotted_cell_index)

            if index in self.unit_drag_lines:
                self.unit_drag_lines[index].hide()

            try:
                self.active_ROI.remove(self.unit_drag_lines[index])
            except ValueError:
                pass
            except KeyError:
                pass

            self.drag_active = False
            self.actions_made = True



import numpy as np
from sklearn.decomposition import PCA


def feature_Energy(data):
    """
    Normlized energy calculation discussed from:
    Quantitative measures of cluster quality for use in extracellular recordings by Redish et al.
    Args:
        data: ndarray representing spike data (num_channels X num_spikes X samples_per_spike)
        iPC (optional): the number
        norm (optional): normalize the waveform (True), or not (False).
    Returns:
        E:
    """
    # energy sqrt of the sum of the squares of each point of the waveform, divided by number of samples in waveform
    # energy and first principal component coefficient
    nSamp = data.shape[-1]

    sum_squared = np.sum(data ** 2, axis=-1)

    if len(np.where(sum_squared < 0)[0]) > 0:
        raise ValueError('summing squared values produced negative number!')

    E = np.divide(np.sqrt(sum_squared), np.sqrt(nSamp))  # shape: channel numbers x spike number
    # E[np.where(E == 0)] = 1  # remove any 0's so we avoid dividing by zero

    return E.T


def feature_WavePCX_scikit(data, iPC=1):

    outputs = np.zeros((data.shape[0], data.shape[1]))
    for i, channel_data in enumerate(data):
        pca = PCA(n_components=iPC)
        principalComponents = pca.fit_transform(channel_data)
        outputs[i, :] = principalComponents[:, iPC-1]

    return outputs.T


def feature_WavePCX(data, iPC=1, norm=False):
    """Creates the principal components for the waveforms
    Args:
        data: ndarray representing spike data (num_channels X num_spikes X samples_per_spike)
        iPC (optional): the number
        norm (optional): normalize the waveform (True), or not (False).
    Returns:
        FD:
    """

    nCh, nSpikes, nSamp = data.shape

    wavePCData = np.zeros((nSpikes, nCh))

    if norm:
        l2norms = np.sqrt(np.sum(data ** 2, axis=-1))
        l2norms = l2norms.reshape((nCh, -1, 1))

        data = np.divide(data, l2norms)

        # removing NaNs from potential division by zero
        data[np.where(np.isnan(data) == True)] = 0

    for i, w in enumerate(data):
        av = np.mean(w, axis=0)  # row wise average
        # covariance matrix
        cv = np.cov(w.T)
        sd = np.sqrt(np.diag(cv)).T  # row wise standard deviation
        pc, _, _, _ = wavePCA(cv)

        # standardize data to zero mean and unit variance
        wstd = (w - av) / (sd)

        # project the data onto principal component axes
        wpc = np.dot(wstd, pc)

        wavePCData[:, i] = wpc[:, iPC - 1]

    return wavePCData


def wavePCA(cv):
    """
    Principal Component Analysis of standardized waveforms from a
    given (unstandardized) waveform covariance matrix cv(nSamp,nSamp).
    Args:
        cv: nSamp x nSamp wavefrom covariance matrix (unnormalized)
    Returns:
        pc: column oriented principal components (Eigenvectors)
        rpc: column oriented Eigenvectors weighted with their relative amplitudes
        ev: eigenvalues of SVD (= std deviation of data projected onto pc)
        rev: relative eigenvalues so that their sum = 1
    """
    sd = np.sqrt(np.diag(cv)).reshape((-1, 1))  # row wise standard deviation

    # standardized covariance matrix
    cvn = np.divide(cv, np.multiply(sd, sd.T))

    u, ev, pc = np.linalg.svd(cvn)

    # the pc is transposed in the matlab version
    pc = pc.T

    ev = ev.reshape((-1, 1))

    rev = ev / np.sum(ev)  # relative eigne values so that their sum = 1

    rpc = np.multiply(pc, rev.T)

    return pc, rpc, ev, rev


def feature_Amplitude(data):
    amplitude = np.amax(data, axis=2) - np.amin(data, axis=2)

    return amplitude.T


def feature_Peak(data):
    peak = np.amax(data, axis=2)

    return peak.T


def feature_Trough(data):
    trough = np.amin(data, axis=2)

    return trough.T


def feature_TroughTime(data):
    troughT = np.argmin(data, axis=2)

    return troughT.T


def feature_PeakTime(data):
    peakT = np.argmax(data, axis=2)

    return peakT.T


def feature_voltageTime(data, time):
    voltageTime = data[:, :, time]

    return voltageTime.T


def CreateFeatures(data, featuresToCalculate=['Energy', 'WavePCX!1']):
    """Creates the features to be analyzed
    Args:
        featuresToCalculate:
    Returns:
        FD:
    """

    FD = np.array([])

    for feature_name in featuresToCalculate:
        if 'WavePCX' in feature_name:
            # instead of creating a function for every PCX you want to use, I just
            # created one. Use WavePCX!1 for PC1, WavePCX!2 for PC2, etc.

            # all the feature functions are named 'feature_featureName'
            # WavePCX is special though since you need a PC number so separate that
            feature_name, pc_number = feature_name.split('!')
            variables = 'data, %s' % pc_number
        elif 'voltageTime' in feature_name:
            # here we will use voltageTime!index_number
            feature_name, time = feature_name.split('!')
            variables = 'data, %s' % time
        else:
            variables = 'data'

        fnc_name = 'feature_%s' % feature_name
        current_FD = eval("%s(%s)" % (fnc_name, variables))

        if len(FD) == 0:
            FD = current_FD
        else:
            FD = np.hstack((FD, current_FD))

        current_FD = None

    return FD



import numpy as np
import time


def get_grid_dimensions(n_cells, method='auto', n=None):
    """
    This function will automate the rows and columns for grid of unit plots. Right now it will
    attempt to just make the grid as square as possible. However, Tint does 5 per row, so I might
    conform to that as well.

    method='auto' will keep it as square of a shape as possible
    method='5per' will make it 5 cells per row
    """

    if method == 'auto':
        # try to make the shape as square as possible, if there are 9 cells it will do a
        # 3 by 3 formation.

        if np.sqrt(n_cells).is_integer():
            rows = int(np.sqrt(n_cells))
            cols = int(rows)
        else:

            value1 = int(np.ceil(np.sqrt(n_cells)))
            value2 = int(np.floor(np.sqrt(n_cells)))

            if value1 * value2 < n_cells:
                value2 = int(np.ceil(np.sqrt(n_cells)))

            cols, rows = sorted(np.array([value1, value2]))

        # I prefer there being more columns than rows if necessary.
        if rows <= cols:
            return rows, cols
        else:
            return cols, rows
    elif method == '5per':
        # return 5 units per row

        cols = 5
        rows = int(np.ceil(n_cells/cols))
        return rows, cols

    elif method == 'nper':
        if n is None:
            raise ValueError('Invalid n value!')

        else:
            cols = n
            rows = int(np.ceil(n_cells / cols))

        return rows, cols


def get_index_from_cell(self, cell):
    for index, value in self.unit_plots.items():
        if len(value) == 1:
            continue
        else:
            if value[1] == cell:
                return index
    return None


def get_index_from_old_cell(cell, unit_plots):
    for index, value in unit_plots.items():
        if len(value) == 1:
            continue
        else:
            if value[1] == cell:
                return index
    return None


def get_cell_from_index(self, index):
    return self.unit_plots[index][1]


def get_channel_y_edges(channel_range=256, n_channels=4):
    return np.arange(n_channels + 1) * -channel_range


def get_channel_from_y(y_value, channel_range=256, n_channels=4):
    """Get the channel to look for the line crossing in"""
    edges = get_channel_y_edges(channel_range=channel_range, n_channels=n_channels)
    for channel in np.arange(n_channels):
        if edges[channel] >= y_value >= edges[channel+1]:
            return channel
    return None


def getSlope(points):
    points_diffs = np.diff(points, axis=0).flatten()
    slope = points_diffs[1] / points_diffs[0]

    if points_diffs[0] == 0:
        print('hi')

    return slope


def findSpikeSubsample(data, max_n):
    if data.shape[0] > max_n:
        # show the clipped values

        ymax = np.amax(data)
        ymin = np.amin(data)

        # remove the clipped cells first
        data_bool = np.zeros_like(data)
        data_bool[np.where((data >= ymax) | (data <= ymin))] = 1
        data_boolsum = np.sum(data_bool, axis=1)
        clipped_cell_i = np.where(data_boolsum >= 5)[0]
        non_clipped_cell_i = np.setdiff1d(np.arange(data.shape[0]), clipped_cell_i.flatten)

        # find the outlying points
        data_i = np.unique(np.concatenate((np.argmax(data[non_clipped_cell_i, :], axis=0),
                                           np.argmin(data[non_clipped_cell_i, :], axis=0))
                                          ).flatten())

        if data_i.shape[0] < max_n:

            remaining_index_choices = np.setdiff1d(np.arange(data.shape[0] - 1), data_i)

            # now just take evenly spaced indices of the remaining choices
            data_i = np.concatenate((data_i,
                                     remaining_index_choices[np.linspace(0,
                                                                         len(remaining_index_choices) - 1,
                                                                         num=(max_n - len(data_i))
                                                                         ).astype(int)]))
        elif data_i.shape[0] > max_n:
            data_i = data_i[:max_n]

        data = data[data_i, :]

    else:
        data_i = np.arange(data.shape[0])
    return data, data_i


def getYIntercept(slope, point):
    return point[1] - slope * point[0]


def find_spikes_crossed(points, unit_data, samples_per_spike=50):
    # calculate the line equation so we can get the points on the line
    slope = getSlope(points)
    y0 = getYIntercept(slope, points[0])

    x_values = np.sort(points[:, 0]).flatten()

    start = x_values[0]; stop = x_values[1] + 1
    if start < 0:
        start = 0

    if stop > samples_per_spike:
        stop = samples_per_spike

    x = np.arange(start, stop)

    cross_line = slope * x + y0  # equation for the user's line

    unit_data_bool = np.intersect1d(x, np.arange(samples_per_spike))

    crossed_cells = np.unique(np.where(np.diff(np.sign(cross_line - unit_data[:, unit_data_bool.astype(int)])))[0])

    return crossed_cells


def setPlotTitle(plot, cell, original_cell_count=None, current_cell_count=None):
    """
    This function will set the title for each of the plots so we know which cell is being plotted.

    :param plot:
    :param cell:
    :param original_cell_count:
    :param current_cell_count:
    :return:
    """
    title_text = "Cell %d, " % cell
    if current_cell_count is not None:
        title_text += "%d spikes" % current_cell_count
    if original_cell_count is not None:
        percent_change = np.abs(100 * (original_cell_count - current_cell_count) / original_cell_count)
        if current_cell_count > original_cell_count:
            title_text += " <span style='color:green'>(add %.2f%%)</span>" % percent_change
        elif current_cell_count < original_cell_count:
            title_text += " <span style='color:red'>(rem %.2f%%)</span>" % percent_change
    plot.setTitle(title_text)


def get_index_from_position(self, position):
    if position in self.unit_positions.keys():
        return self.unit_positions[position]
    return None


def get_old_index_from_position(self, position):
    if position in self.old_positions.keys():
        return self.old_positions[position]
    return None


def get_position_from_index(self, index):
    for position, index_ in self.unit_positions.items():
        if index == index_:
            return position
    return None


def add_graph_limits():
    pass


def validateMoveValue(channel_value):
    """
    Determine if the channel value is valid or not
    :param channel_value:
    :return:
    """
    if channel_value < 0:
        return False
    elif channel_value > 30:
        return False
    return True


def validateMaxSpikes(max_spikes):
    """
    :param channel_value:
    :return:
    """
    try:
        max_spike = int(max_spikes)
        if max_spike < 0:
            return False
    except:
        return False

    return True


def get_max_spikes(self):
    max_spike_plots = self.max_spike_plots_text.text()
    max_spike_valid = validateMaxSpikes(max_spike_plots)
    if max_spike_valid:
        max_spike_plots = int(max_spike_plots)
    else:
        self.choice = None
        self.LogError.signal.emit('invalidMaxSpikes')
        while self.choice is None:
            time.sleep(0.1)
        return None

    return max_spike_plots


def get_next_action(self):
    return max(self.latest_actions.keys()) + 1


def clear_unit(self, cell):
    """
    This function will clear the plots for a cell that has been removed.

    :param self:
    :param cell:
    :return:
    """

    if cell == 0:
        return

    index = get_index_from_cell(self, cell)

    for chan in np.arange(self.n_channels):
        self.unit_plots[index][0].removeItem(self.plot_lines[index][chan])
        self.unit_plots[index][0].removeItem(self.avg_plot_lines[index][chan])


def moveToChannel(self, origin):

    if origin == 'main':
        for key in self.PopUpCutWindow.keys():
            self.PopUpCutWindow[key].move_to_channel.setText(self.move_to_channel.text())
    elif origin == 'popup':

        cell = self.cell
        # for key in self.
        self.mainWindow.move_to_channel.setText(self.move_to_channel.text())


def maxSpikesChange(self, origin):

    if origin == 'main':
        for key in self.PopUpCutWindow.keys():
            self.PopUpCutWindow[key].max_spike_plots_text.setText(self.max_spike_plots_text.text())
    elif origin == 'popup':
        cell = self.cell
        self.mainWindow.max_spike_plots_text.setText(self.max_spike_plots_text.text())
    self.max_spike_plots = None  # we will set this value to none as to re-validate the value



from __future__ import absolute_import, division, print_function

import sys
import os
sys.path.append(os.path.dirname(__file__))

from .core.custom_widgets import *
from .core.default_parameters import *
from .core.feature_plot import *
from .core.feature_functions import *
from .core.gui_utils import *
from .core.plot_functions import *
from .core.plot_utils import *
from .core.PopUpCutting import *
from .core.Tint_Matlab import *
from .core.undo import *
from .core.waveform_cut_functions import *
from .core.writeCut import *

__all__ = ['core', 'main', 'exporters']

# __path__ = __import__('pkgutil').extend_path(__path__, __name__)





from PyQt5 import QtWidgets, QtGui, QtCore
import pyqtgraph as pg
import os
from .default_parameters import project_name, channel_range, max_spike_plots, default_move_channel, max_num_actions
from .plot_utils import CustomViewBox, MultiLine, get_channel_color
import numpy as np
from .gui_utils import Communicate
from functools import partial
from .plot_functions import replot_unit, reconfigure_units
from .waveform_cut_functions import get_channel_from_y, get_channel_y_edges, findSpikeSubsample, \
    get_index_from_cell, get_max_spikes, setPlotTitle, find_spikes_crossed, validateMoveValue, clear_unit, \
    maxSpikesChange, moveToChannel, get_next_action
from .undo import undo_function
import time


class PopUpCutWindow(QtWidgets.QWidget):
    def __init__(self, mainWindow):
        super(PopUpCutWindow, self).__init__()

        self.mainWindow = mainWindow

        pg.setConfigOptions(antialias=True)

        QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create('GTK+'))

        # declaring the icon image
        self.setWindowIcon(QtGui.QIcon(os.path.join(self.mainWindow.IMG_DIR, 'GEBA_Logo.png')))

        self.setWindowTitle("%s - Popup Cutting Window" % project_name)  # sets the main window title

        self.channel_win = None
        self.unit_win = None
        self.hide_btn = None
        self.channel_number = None

        self.drag_active = False

        self.choice = None
        self.LogError = Communicate()
        self.LogError.signal.connect(self.raiseError)

        self.active_ROI = []

        self.avg_plot_lines = {}
        self.plot_lines = {}

        self.channel_avg_plot_lines = {}
        self.channel_plot_lines = {}

        self.index = None
        self.cell = None

        self.PopUpActive = False

        self.samples_per_spike = None
        self.n_channels = None

        self.initialize()  # initializes the main window

    def raiseError(self, error):

        if 'InvalidMoveChannel' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Move to Cell Value!",
                                                         "The value you have chosen for the 'Move to Cell' value is "
                                                         "invalid, please choose a valid value before continuing!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'SameChannelInvalid' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Same Channel Error!",
                                                         "The value you have chosen for the 'Move to Cell' value is "
                                                         "the same as the cell you are cutting from! If you would like "
                                                         "to move these selected spikes to a different channel, please "
                                                         "choose another channel!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'ActionsMade' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Are you sure...?",
                                                         "You have performed some actions that will be lost when you"
                                                         " reload this cut file. Are you sure you want to continue?",
                                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        elif 'invalidMaxSpikes' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Max Spikes",
                                                         "The number chosen for Max Spikes is invalid!",
                                                         QtWidgets.QMessageBox.Ok)

    def initialize(self):

        # parameter layout

        channel_layout = QtWidgets.QHBoxLayout()
        channel_label = QtWidgets.QLabel("Channel:")
        self.channel_number = QtWidgets.QComboBox()
        self.channel_number.currentIndexChanged.connect(self.plot_channel)

        channel_layout.addWidget(channel_label)
        channel_layout.addWidget(self.channel_number)

        max_spike_label = QtWidgets.QLabel("Max Plot Spikes:")
        self.max_spike_plots_text = QtWidgets.QLineEdit()
        self.max_spike_plots_text.setToolTip("This is the maximum number of spikes to plot.")
        self.max_spike_plots_text.setText(self.mainWindow.max_spike_plots_text.text())
        self.max_spike_plots_text.textChanged.connect(lambda: maxSpikesChange(self, 'popup'))

        max_spikes_layout = QtWidgets.QHBoxLayout()
        max_spikes_layout.addWidget(max_spike_label)
        max_spikes_layout.addWidget(self.max_spike_plots_text)

        move_to_layout = QtWidgets.QHBoxLayout()

        move_to_label = QtWidgets.QLabel("Move to Cell:")
        self.move_to_channel = QtWidgets.QLineEdit()
        self.move_to_channel.setText(self.mainWindow.move_to_channel.text())
        self.move_to_channel.textChanged.connect(lambda: moveToChannel(self, 'popup'))

        move_to_layout.addWidget(move_to_label)
        move_to_layout.addWidget(self.move_to_channel)

        parameter_layout = QtWidgets.QHBoxLayout()
        parameter_layout.addStretch(1)
        for object_ in [channel_layout, move_to_layout, max_spikes_layout]:
            if 'Layout' in object_.__str__():
                parameter_layout.addLayout(object_)
                parameter_layout.addStretch(1)
            else:
                parameter_layout.addWidget(object_, 0, QtCore.Qt.AlignCenter)
                parameter_layout.addStretch(1)

        # plot layout
        self.channel_win = pg.GraphicsWindow()
        self.channel_plot = self.channel_win.addPlot(row=0, col=0, viewBox=CustomViewBox(self, self.channel_win))
        self.channel_plot.hideAxis('left')  # remove the y-axis
        self.channel_plot.hideAxis('bottom')  # remove the x axis
        self.channel_plot.hideButtons()  # hide the auto-resize button
        self.channel_plot.setMouseEnabled(x=False, y=False)  # disables the mouse interactions
        self.channel_plot.enableAutoRange(False, False)
        self.vb_channel_plot = self.channel_plot.vb
        self.vb_channel_plot.mouseDragEvent = partial(dragPopup, self, 'channel',
                                                   self.vb_channel_plot)  # overriding the drag event
        self.vb_channel_plot.mouseClickEvent = partial(mouse_click_eventPopup, self, self.vb_channel_plot)

        self.channel_drag_lines = None

        self.unit_win = pg.GraphicsWindow()
        self.unit_plot = self.unit_win.addPlot(row=0, col=0, viewBox=CustomViewBox(self, self.unit_win))
        self.unit_plot.hideAxis('left')  # remove the y-axis
        self.unit_plot.hideAxis('bottom')  # remove the x axis
        self.unit_plot.hideButtons()  # hide the auto-resize button
        self.unit_plot.setMouseEnabled(x=False, y=False)  # disables the mouse interactions
        self.unit_plot.enableAutoRange(False, False)

        self.vb_unit_plot = self.unit_plot.vb

        self.vb_unit_plot.mouseDragEvent = partial(dragPopup, self, 'unit', self.vb_unit_plot)  # overriding the drag event
        self.vb_unit_plot.mouseClickEvent = partial(mouse_click_eventPopup, self, self.vb_unit_plot)

        self.unit_drag_lines = None

        plot_layout = QtWidgets.QHBoxLayout()
        for _object in [self.channel_win, self.unit_win]:
            plot_layout.addWidget(_object)

        # button layout

        button_layout = QtWidgets.QHBoxLayout()

        self.undo_btn = QtWidgets.QPushButton("Undo")
        self.undo_btn.clicked.connect(lambda: undo_function(self.mainWindow))

        self.hide_btn = QtWidgets.QPushButton("Hide")
        self.hide_btn.clicked.connect(self.hideF)

        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        self.cancel_btn.clicked.connect(self.cancel)

        for widget in [self.undo_btn, self.hide_btn, self.cancel_btn]:
            button_layout.addWidget(widget)

        # adding all widgets/layouts to the main layout

        layout_order = [parameter_layout, plot_layout, button_layout]
        add_Stretch = [False, False, False]
        # ---------------- add all the layouts and widgets to the Main Window's layout ------------ #
        popupLayout = QtWidgets.QVBoxLayout()

        for widget, addStretch in zip(layout_order, add_Stretch):
            if 'Layout' in widget.__str__():
                popupLayout.addLayout(widget)
                if addStretch:
                    popupLayout.addStretch(1)
            else:
                popupLayout.addWidget(widget, 0, QtCore.Qt.AlignCenter)
                if addStretch:
                    popupLayout.addStretch(1)

        self.setLayout(popupLayout)  # defining the layout of the Main Window

    def cancel(self):
        self.hideF()
        self.mainWindow.PopUpCutWindow.pop(self.cell)

    def hideF(self):
        self.hide()
        self.PopUpActive = False

    def plot(self, index, cell):

        self.PopUpActive = True

        self.index = index
        self.cell = cell

        if self.n_channels is None:
            self.n_channels = self.mainWindow.n_channels
            for channel in np.arange(self.n_channels):
                self.channel_number.addItem(str(int(channel+1)))
            self.channel_number.setCurrentIndex(0)
            self.unit_plot.setYRange(0, -self.n_channels * channel_range, padding=0)

        if self.samples_per_spike is None:
            self.samples_per_spike = self.mainWindow.samples_per_spike
            self.unit_plot.setXRange(0, self.samples_per_spike, padding=0)  # set the x-range
            self.channel_plot.setXRange(0, self.samples_per_spike, padding=0)  # set the x-range

        setTitle = False
        for channel in np.arange(self.n_channels):

            if index in self.mainWindow.unit_data.keys():
                plot_data = self.mainWindow.unit_data[index][channel]
            else:
                # self.reset_plots()
                self.cancel()
                return

            plot_data_avg = np.mean(plot_data, axis=0).reshape((1, -1))

            current_n = plot_data.shape[0]

            plot_data = plot_data[self.mainWindow.cell_subsample_i[cell][channel], :]

            if not setTitle:
                if cell in self.mainWindow.original_cell_count.keys():
                    setPlotTitle(self.unit_plot, cell, original_cell_count=self.mainWindow.original_cell_count[cell],
                                 current_cell_count=current_n)
                else:
                    setPlotTitle(self.unit_plot, cell, current_cell_count=current_n)
                setTitle = True

            if index not in self.plot_lines.keys():
                self.plot_lines[index] = {
                    channel: MultiLine(np.tile(np.arange(self.samples_per_spike), (plot_data.shape[0], 1)),
                                       plot_data, pen_color=get_channel_color(cell))}
            else:
                if channel in self.plot_lines[index].keys():
                    self.unit_plot.removeItem(self.plot_lines[index][channel])
                self.plot_lines[index][channel] = MultiLine(np.tile(np.arange(self.samples_per_spike),
                                                                     (plot_data.shape[0], 1)),  plot_data,
                                                            pen_color=get_channel_color(cell))

            if index not in self.avg_plot_lines.keys():
                self.avg_plot_lines[index] = {channel: MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                                                 plot_data_avg, pen_color='w', pen_width=2)}
            else:
                if channel in self.avg_plot_lines[index].keys():
                    self.unit_plot.removeItem(self.avg_plot_lines[index][channel])

                self.avg_plot_lines[index][channel] = MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                                                plot_data_avg, pen_color='w', pen_width=2)

            self.unit_plot.addItem(self.plot_lines[index][channel])
            self.unit_plot.addItem(self.avg_plot_lines[index][channel])

        self.plot_channel()

        self.show()

    def plot_channel(self):

        self.channel_plot.clear()

        if self.samples_per_spike is None:
            return

        channel = int(self.channel_number.currentText()) - 1

        plot_data = self.mainWindow.unit_data[self.index][channel]
        plot_data_avg = np.mean(plot_data, axis=0).reshape((1, -1))
        plot_data = plot_data[self.mainWindow.cell_subsample_i[self.cell][channel], :]

        if self.index not in self.channel_plot_lines.keys():
            self.channel_plot_lines[self.index] = {
                channel: MultiLine(np.tile(np.arange(self.samples_per_spike), (plot_data.shape[0], 1)),
                                   plot_data, pen_color=get_channel_color(self.cell))}
        else:
            if channel in self.channel_plot_lines[self.index].keys():
                self.channel_plot.removeItem(self.channel_plot_lines[self.index][channel])

            self.channel_plot_lines[self.index][channel] = MultiLine(np.tile(np.arange(self.samples_per_spike),
                                                                (plot_data.shape[0], 1)), plot_data,
                                                        pen_color=get_channel_color(self.cell))

        if self.index not in self.channel_avg_plot_lines.keys():
            self.channel_avg_plot_lines[self.index] = {channel: MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                                             plot_data_avg, pen_color='w', pen_width=2)}
        else:
            if channel in self.channel_avg_plot_lines[self.index].keys():
                self.channel_plot.removeItem(self.channel_avg_plot_lines[self.index][channel])

            self.channel_avg_plot_lines[self.index][channel] = MultiLine(np.arange(self.samples_per_spike).reshape((1, -1)),
                                                            plot_data_avg, pen_color='w', pen_width=2)

        ymin, ymax = get_ylimits(channel, channel_range=channel_range, n_channels=self.n_channels)

        self.channel_plot.addItem(self.channel_plot_lines[self.index][channel])
        self.channel_plot.addItem(self.channel_avg_plot_lines[self.index][channel])

        self.channel_plot.setYRange(ymin, ymax, padding=0)

    def reset_plots(self):
        self.unit_plot.clear()
        self.unit_plot.setTitle('')
        self.channel_plot.clear()

    def reset_data(self):

        self.reset_plots()

        self.cell = None
        self.index = None

        self.unit_drag_lines = None
        self.channel_drag_lines = None

        self.drag_active = False

        self.active_ROI = []
        self.plot_lines = {}
        self.avg_plot_lines = {}

        self.channel_avg_plot_lines = {}
        self.channel_plot_lines = {}


def get_ylimits(channel, channel_range=256, n_channels=4):
    edges = get_channel_y_edges(channel_range=channel_range, n_channels=n_channels)
    return edges[channel+1], edges[channel]


def dragPopup(self, mode, vb, ev=None):
    # global vb, lr
    if ev.button() == QtCore.Qt.LeftButton:

        # defining the start of the selected region
        points = [[vb.mapToView(ev.buttonDownPos()).x(),
                   vb.mapToView(ev.buttonDownPos()).y()],
                  [vb.mapToView(ev.pos()).x(),
                   vb.mapToView(ev.pos()).y()]]

        if mode == 'unit':
            self.unit_plot.removeItem(self.unit_drag_lines)
            self.unit_drag_lines = pg.PolyLineROI(points)
            self.unit_plot.addItem(self.unit_drag_lines)
            self.active_ROI = [self.unit_drag_lines]
            self.drag_active = True
        elif mode == 'channel':
            self.channel_plot.removeItem(self.channel_drag_lines)
            self.channel_drag_lines = pg.PolyLineROI(points)
            self.channel_plot.addItem(self.channel_drag_lines)
            self.active_ROI = [self.channel_drag_lines]
            self.drag_active = True

        ev.accept()
    else:
        pg.ViewBox.mouseDragEvent(vb, ev)


def mouse_click_eventPopup(self, vb, ev=None):

    if ev.button() == QtCore.Qt.RightButton:
        # open menu
        pg.ViewBox.mouseClickEvent(vb, ev)

    elif ev.button() == QtCore.Qt.LeftButton:

        # hopefully drag event
        pg.ViewBox.mouseClickEvent(vb, ev)

    elif ev.button() == QtCore.Qt.MiddleButton:
        # then we will accept the changes

        if self.index is None:
            return

        if self.unit_drag_lines in self.active_ROI or self.channel_drag_lines in self.active_ROI:
            # then we have an active ROI
            # we will get the x,y positions (rounded to the nearest int) of the selected line

            try:
                invalid_cell_number = int(self.move_to_channel.text())
            except:
                self.choice = None
                self.LogError.signal.emit('InvalidMoveChannel')
                while self.choice is None:
                    time.sleep(0.1)
                return

            if not validateMoveValue(invalid_cell_number):
                self.choice = None
                self.LogError.signal.emit('InvalidMoveChannel')
                while self.choice is None:
                    time.sleep(0.1)
                return

            if invalid_cell_number == self.cell:
                self.choice = None
                self.LogError.signal.emit('SameChannelInvalid')
                while self.choice is None:
                    time.sleep(0.1)
                return

            if self.unit_drag_lines in self.active_ROI:
                points = np.rint(np.asarray(self.unit_drag_lines.getState()['points']))
                channel = get_channel_from_y(points[0, 1], channel_range=channel_range, n_channels=self.n_channels)
            elif self.channel_drag_lines in self.active_ROI:
                points = np.rint(np.asarray(self.channel_drag_lines.getState()['points']))
                channel = int(self.channel_number.currentText()) - 1

            max_spikes_changed = False
            if self.mainWindow.max_spike_plots is None:
                max_spike_plots = get_max_spikes(self.mainWindow)
                if max_spike_plots is None:
                    return
                else:
                    self.mainWindow.max_spike_plots = max_spike_plots

                    # update the max spikes by re-calculating the subsample index values
                    for plotted_cell in self.mainWindow.cell_subsample_i.keys():
                        plotted_channel_index = get_index_from_cell(self.mainWindow, plotted_cell)

                        if any(plotted_cell == cell_value for cell_value in [0]):
                            # skip the invalid channel (done later) and the dummy channel (not plotted)
                            continue

                        for data_chan in np.arange(self.mainWindow.n_channels):
                            _, subsample_i = findSpikeSubsample(self.mainWindow.unit_data[plotted_channel_index][data_chan],
                                                                self.mainWindow.max_spike_plots)

                            if plotted_cell not in self.mainWindow.cell_subsample_i.keys():
                                self.mainWindow.cell_subsample_i[plotted_cell] = {data_chan: subsample_i}
                            else:
                                self.mainWindow.cell_subsample_i[plotted_cell][data_chan] = subsample_i
                    max_spikes_changed = True

            unit_data = self.mainWindow.unit_data[self.index][channel]
            crossed_cells = find_spikes_crossed(points, unit_data, samples_per_spike=self.samples_per_spike)

            invalid_index = get_index_from_cell(self.mainWindow, invalid_cell_number)
            # remove these spikes from all the channels
            reconfigure = False

            if len(crossed_cells) != 0:
                for data_chan in np.arange(self.mainWindow.n_channels):
                    # append the crossed lines to the invalid cell's plot
                    if invalid_index is not None:
                        # get the invalid data
                        invalid_cell_data = self.mainWindow.unit_data[self.index][data_chan][crossed_cells, :]
                        # update the invalid_data channel with this current data
                        self.mainWindow.unit_data[invalid_index][data_chan] = np.vstack((
                            self.mainWindow.unit_data[invalid_index][data_chan], invalid_cell_data))

                        # update the plotted subsample as well
                        _, subsample_i = findSpikeSubsample(self.mainWindow.unit_data[invalid_index][data_chan], self.mainWindow.max_spike_plots)
                        if invalid_cell_number not in self.mainWindow.cell_subsample_i.keys():
                            self.mainWindow.cell_subsample_i[invalid_cell_number] = {data_chan: subsample_i}
                        else:
                            self.mainWindow.cell_subsample_i[invalid_cell_number][data_chan] = subsample_i
                    else:
                        reconfigure = True

                    self.mainWindow.unit_data[self.index][data_chan] = np.delete(
                        self.mainWindow.unit_data[self.index][data_chan], crossed_cells, axis=0)

                    # recalculate subplot for the channel that the spikes were removed from
                    if len(self.mainWindow.unit_data[self.index][data_chan]) > 0:
                        _, subsample_i = findSpikeSubsample(self.mainWindow.unit_data[self.index][data_chan], self.mainWindow.max_spike_plots)
                        if self.cell not in self.mainWindow.cell_subsample_i.keys():
                            self.mainWindow.cell_subsample_i[self.cell] = {data_chan: subsample_i}
                        else:
                            self.mainWindow.cell_subsample_i[self.cell][data_chan] = subsample_i
                    else:
                        # there is no data left, don't need to worry about the subsampling anymore
                        if self.cell in self.mainWindow.cell_subsample_i.keys():
                            self.mainWindow.cell_subsample_i.pop(self.cell)
                            reconfigure = True

                # check if the cell still exists
                for key, value in self.mainWindow.unit_data[self.index].items():
                    if len(value) == 0:
                        if self.index != -1:
                            self.mainWindow.unit_data.pop(self.index)
                            reconfigure = True
                            break

                # update the bool
                cell_indices = self.mainWindow.cell_indices[self.cell]

                # append invalid cells to the new cell number
                invalid_cells = cell_indices[crossed_cells]
                self.mainWindow.cell_indices[self.cell] = np.delete(cell_indices, crossed_cells)

                # check if there are still indices for this cell, if empty we will remove
                if len(self.mainWindow.cell_indices[self.cell]) == 0:
                    reconfigure = True
                    clear_unit(self.mainWindow, self.cell)  # delete the cell's plots
                    if self.cell in self.mainWindow.original_cell_count.keys():
                        self.mainWindow.original_cell_count.pop(self.cell)
                    if self.cell != 0:
                        self.mainWindow.cell_indices.pop(self.cell)

                if invalid_cell_number in self.mainWindow.cell_indices.keys():
                    # the cell has existed already within the main window, we can just add to this plot
                    self.mainWindow.cell_indices[invalid_cell_number] = np.concatenate((
                        self.mainWindow.cell_indices[invalid_cell_number], invalid_cells))
                else:
                    # this cell is not already plotted, have to add the plot and possibly reconfigure
                    self.mainWindow.cell_indices[invalid_cell_number] = invalid_cells

                    if invalid_cell_number != 0:
                        reconfigure = True

                # add the latest action
                if len(self.mainWindow.latest_actions) == 0 or max_num_actions == 1:
                    self.mainWindow.latest_actions = {0: {'action': 'cut', 'fromCell': self.cell,
                                                          'toCell': invalid_cell_number, 'movedCutIndices': invalid_cells}}
                else:
                    next_action = get_next_action(self.mainWindow)
                    self.mainWindow.latest_actions[next_action] = {'action': 'cut', 'fromCell': self.cell,
                                                                   'toCell': invalid_cell_number,
                                                                   'movedCutIndices': invalid_cells}

            # re-plot on popup
            self.plot(self.index, self.cell)
            for roi in self.active_ROI:
                roi.hide()

            # re-plot the main Window
            if not reconfigure:
                # update plots for the invalid cell and the cell you removed these spikes from
                # no need to reconfigure
                replot_unit(self.mainWindow, self.index)
                invalid_index = get_index_from_cell(self.mainWindow, invalid_cell_number)
                if invalid_index != -1:
                    replot_unit(self.mainWindow, invalid_index)
            else:
                # we will need to reconfigure the main window possibly, do so

                if self.cell in self.mainWindow.cell_indices.keys():
                    replot_unit(self.mainWindow, self.index)
                unique_cells = np.asarray(list(self.mainWindow.cell_indices.keys()))
                reconfigure_units(self.mainWindow, list(unique_cells[unique_cells != 0]))

            if max_spikes_changed:
                for plotted_cell in self.mainWindow.cell_indices.keys():
                    if plotted_cell == invalid_cell_number or plotted_cell == self.cell:
                        continue

                    plotted_cell_index = get_index_from_cell(self.mainWindow, plotted_cell)
                    replot_unit(self.mainWindow, plotted_cell_index)

            self.drag_active = False
            self.mainWindow.actions_made = True



from PyQt5 import QtWidgets, QtCore
from .default_parameters import default_filename
import time
import os

Large_Font = ("Verdana", 12)  # defines two fonts for different purposes (might not be used
Small_Font = ("Verdana", 8)


def center(self):
    """
    centers the window on the screen
    """
    frameGm = self.frameGeometry()
    screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())
    centerPoint = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
    frameGm.moveCenter(centerPoint)
    self.move(frameGm.topLeft())


class Worker(QtCore.QObject):
    """
    This Worker class is for threading using PyQt
    """
    def __init__(self, function, *args, **kwargs):
        super(Worker, self).__init__()
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.start.connect(self.run)

    start = QtCore.pyqtSignal(str)

    @QtCore.pyqtSlot()
    def run(self):
        self.function(*self.args, **self.kwargs)


class Communicate(QtCore.QObject):
    """
    A custom pyqtsignal so that errors and popups can be called from the threads
    to the main window
    """
    signal = QtCore.pyqtSignal(str)


def validate_session(self, set_filename):

    if set_filename == default_filename:
        self.choice = None
        self.LogError.signal.emit('ChooseSession')
        while self.choice is None:
            time.sleep(0.1)
        return False, True
    return True, False


def validate_cut(self, set_filename, cut_filename):
    set_basename = os.path.basename(set_filename)

    if set_basename in cut_filename:
        return True
    else:
        return False
    return True



import numpy as np
from .waveform_cut_functions import get_index_from_cell, get_max_spikes, findSpikeSubsample
from .plot_functions import replot_unit, reconfigure_units


def undo_function(self):

    if len(self.latest_actions) == 0:
        return

    latest_action_key = max(self.latest_actions.keys())

    latest_action = self.latest_actions[latest_action_key]

    if latest_action['action'] == 'cut':
        # here we will have to undo a movement of spikes from one cell to another
        fromCell = latest_action['fromCell']  # the spikes where moved from here originally
        toCell = latest_action['toCell']  # the spikes were moved here in this action
        cut_indices = latest_action['movedCutIndices']  # the cut file indices that were moved in the transfer

        # identify the spikes that we moved to the new cell
        movedBool = np.where(np.isin(self.cell_indices[toCell], cut_indices))[0]

        toCellIndex = get_index_from_cell(self, toCell)
        fromCellIndex = get_index_from_cell(self, fromCell)

        if self.max_spike_plots is None:
            max_spike_plots = get_max_spikes(self)
            if max_spike_plots is None:
                return
            else:
                self.max_spike_plots = max_spike_plots

        reconfigure = False
        for data_chan in np.arange(self.n_channels):
            # add the spike data back to it's original channel

            # get the moved data
            moved_cell_data = self.unit_data[toCellIndex][data_chan][movedBool, :]

            if fromCellIndex is not None:
                # put this data back into the fromCell
                self.unit_data[fromCellIndex][data_chan] = np.vstack((self.unit_data[fromCellIndex][data_chan],
                                                                      moved_cell_data))

                # update the plotted subsample as well
                _, subsample_i = findSpikeSubsample(self.unit_data[fromCellIndex][data_chan], self.max_spike_plots)

                if fromCell not in self.cell_subsample_i.keys():
                    self.cell_subsample_i[fromCell] = {data_chan: subsample_i}
                else:
                    self.cell_subsample_i[fromCell][data_chan] = subsample_i
            else:
                # we don't need to worry about this because the reconfigure function will take care of it
                reconfigure = True

            # remove the data from the channel it was original moved to
            self.unit_data[toCellIndex][data_chan] = np.delete(self.unit_data[toCellIndex][data_chan],
                                                               movedBool, axis=0)

            if len(self.unit_data[toCellIndex][data_chan]) > 0:
                # update the toCell subsample_i as well
                _, subsample_i = findSpikeSubsample(self.unit_data[toCellIndex][data_chan], self.max_spike_plots)
                if toCell not in self.cell_subsample_i.keys():
                    self.cell_subsample_i[toCell] = {data_chan: subsample_i}
                else:
                    self.cell_subsample_i[toCell][data_chan] = subsample_i
            else:
                # there is no data left, don't need to worry about the sub-sampling anymore
                if toCell in self.cell_subsample_i.keys():
                    self.cell_subsample_i.pop(toCell)
                    reconfigure = True

        # this cell no longer exists
        for key, value in self.unit_data[toCellIndex].items():
            if len(value) == 0:
                if toCellIndex != -1:
                    # -1 is the index of the dummy cell (0), lets make sure not to remove this
                    self.unit_data.pop(toCellIndex)
                    reconfigure = True
                    break

        # make sure that the spike indices are also moved back to their original cell
        cell_indices = self.cell_indices[toCell]
        undo_cells = cell_indices[movedBool]
        self.cell_indices[toCell] = np.delete(cell_indices, movedBool)

        # remove the cell indices key for this cell if it has no data anymore
        if len(self.cell_indices[toCell]) == 0:
            reconfigure = True
            if toCell != 0:
                self.cell_indices.pop(toCell)

        # add the spikes back to where they used to be
        # determine if we need to reconfigure the main window
        if fromCell in self.cell_indices.keys():
            # the cell has existed already within the main window, we can just add to this plot
            self.cell_indices[fromCell] = np.concatenate((self.cell_indices[fromCell], undo_cells))
        else:
            # this cell is not already plotted, have to add the plot and possibly reconfigure
            self.cell_indices[fromCell] = undo_cells
            reconfigure = True

        # plot the data
        if not reconfigure:
            # update plots for the invalid cell and the cell you removed these spikes from
            # no need to reconfigure
            if toCellIndex != -1:
                replot_unit(self, toCellIndex)

            replot_unit(self, fromCellIndex)
        else:
            # we will need to reconfigure the main window possibly, do so
            if fromCellIndex in self.unit_plots.keys():
                replot_unit(self, fromCellIndex)
            unique_cells = np.asarray(list(self.cell_indices.keys()))
            reconfigure_units(self, list(unique_cells[unique_cells != 0]))

        # remove this action from the dictionary
        self.latest_actions.pop(latest_action_key)

    else:
        print('The following action has not been coded yet: %s' % latest_action['action'])



project_name = 'gebaSpike'

default_filename = 'Choose a filename!'

defaultXAxis = 'PC1'
defaultYAxis = 'PC2'
defaultZAxis = 'PC3'

# openGL = False
openGL = True

gridLines = True

feature_spike_size = 2
feature_spike_opacity = 0.5

# unitMode = 'MatplotWidget'
unitMode = 'PyQt'

default_move_channel = str(0)

channel_range = 256

max_spike_plots = 2000

max_num_actions = 5  # max number of actions to remember



import sys
import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets
from pyqtgraph.widgets.MatplotlibWidget import MatplotlibWidget
from core.gui_utils import validate_session, Communicate, validate_cut
from core.default_parameters import project_name, default_filename, defaultXAxis, defaultYAxis, defaultZAxis, openGL, \
    default_move_channel, max_spike_plots
from core.Tint_Matlab import find_tetrodes
from core.plot_functions import plot_session
from core.waveform_cut_functions import moveToChannel, maxSpikesChange
from core.undo import undo_function
from core.feature_plot import feature_name_map
from core.PopUpCutting import PopUpCutWindow
from core.writeCut import write_cut, write_clu
# from core.plot_functions import get_index_from_cell
import pyqtgraph.opengl as gl
import os
import json
import time
import numpy as np


class MainWindow(QtWidgets.QWidget):

    def __init__(self):
        super(MainWindow, self).__init__()

        self.setWindowTitle("%s - Main Window" % project_name)  # sets the main window title

        # initializing attributes
        self.cut_filename = None
        self.choose_cut_filename_btn = None
        self.feature_win = None
        self.quit_btn = None
        self.filename = None
        self.choose_filename_btn = None
        self.x_axis_cb = None
        self.y_axis_cb = None
        self.z_axis_cb = None
        self.tetrode_cb = None
        self.choice = None
        self.plot_btn = None
        self.feature_plot = None

        self.latest_actions = {}

        self.actions_made = False

        self.drag_active = False

        self.last_drag_index = None

        self.feature_data = None
        self.tetrode_data = None
        self.cut_data = None
        self.cut_data_original = None
        self.spike_times = None
        self.scatterItem = None
        self.glViewWidget = None
        self.feature_plot_added = False
        self.samples_per_spike = None

        self.unit_positions = {}
        self.cell_subsample_i = {}
        self.unit_drag_lines = {}
        self.active_ROI = []
        self.unit_data = {}

        self.xline = None
        self.yline = None
        self.zline = None

        self.max_spike_plots = None

        self.n_channels = None

        self.invalid_channel = None

        self.cell_indices = {}

        self.spike_colors = None

        self.original_cell_count = {}

        self.unit_plots = {}
        self.vb = {}
        self.plot_lines = {}
        self.avg_plot_lines = {}
        self.unit_rows = 0
        self.unit_cols = 0

        self.LogError = Communicate()
        self.LogError.signal.connect(self.raiseError)

        # various setting files
        project_dir = os.path.dirname(os.path.abspath("__file__"))
        if os.path.basename(project_dir) != project_name:
            project_dir = os.path.dirname(sys.argv[0])

        # defining the directory filepath
        self.PROJECT_DIR = project_dir  # project directory
        self.SETTINGS_DIR = os.path.join(self.PROJECT_DIR, 'settings')  # settings directory

        self.settings_filename = os.path.join(self.SETTINGS_DIR, 'settings.json')
        self.settings = self.get_settings()

        self.IMG_DIR = os.path.join(self.PROJECT_DIR, 'img')  # settings directory

        # create Settings Directory
        if not os.path.exists(self.SETTINGS_DIR):
            os.mkdir(self.SETTINGS_DIR)

        # create settings file
        if not os.path.exists(self.settings_filename):
            with open(self.settings_filename, 'w') as f:
                json.dump({}, f)

        # setting PyQtGraph settings
        # pg.setConfigOption('background', 'w')
        # pg.setConfigOption('foreground', 'k')
        pg.setConfigOptions(antialias=True)

        self.PopUpCutWindow = {}
        # self.PopUpCutWindow = PopUpCutWindow(self)

        QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create('GTK+'))

        self.setWindowIcon(QtGui.QIcon(os.path.join(self.IMG_DIR, 'GEBA_Logo.png')))  # declaring the icon image
        self.initialize()  # initializes the main window

    def initialize(self):
        """
        This method will initialize the Main Window, create all the widgets and what not
        :return:
        """

        # ---------------- session filename options ------------------------------------------

        filename_grid_layout = QtWidgets.QGridLayout()

        filename_label = QtWidgets.QLabel("Filename:")
        self.filename = QtWidgets.QLineEdit()
        self.filename.setText(default_filename)
        self.filename.textChanged.connect(self.filename_changed)
        self.filename.setToolTip('The name of the session that you will analyze in gebaSpike!')
        self.choose_filename_btn = QtWidgets.QPushButton("Choose Filename")
        self.choose_filename_btn.setToolTip('This button will allow you to choose the filename to analyze!')
        self.choose_filename_btn.clicked.connect(self.choose_filename)

        filename_grid_layout.addWidget(self.choose_filename_btn, *(0, 0))
        filename_grid_layout.addWidget(filename_label, *(0, 1), QtCore.Qt.AlignRight)

        # ------------------- cut filename options ---------------------------------

        cut_filename_label = QtWidgets.QLabel("Cut Filename:")
        self.cut_filename = QtWidgets.QLineEdit()
        self.cut_filename.setText(default_filename)
        self.cut_filename.setToolTip('The name of the cut file containing the sorted values!')
        self.choose_cut_filename_btn = QtWidgets.QPushButton("Choose Cut Filename")
        self.choose_cut_filename_btn.setToolTip('This button will allow you to choose a cut file to use!')
        self.choose_cut_filename_btn.clicked.connect(self.choose_cut_filename)

        line_edit_layout = QtWidgets.QVBoxLayout()
        line_edit_layout.addWidget(self.filename)
        line_edit_layout.addWidget(self.cut_filename)

        filename_grid_layout.addWidget(self.choose_cut_filename_btn, *(1, 0))
        filename_grid_layout.addWidget(cut_filename_label, *(1, 1), QtCore.Qt.AlignRight)

        filename_grid_layout.addLayout(line_edit_layout, *(0, 2), 2, 1)

        # -----------

        max_spike_label = QtWidgets.QLabel("Max Plot Spikes:")
        self.max_spike_plots_text = QtWidgets.QLineEdit()
        self.max_spike_plots_text.setToolTip("This is the maximum number of spikes to plot.")
        self.max_spike_plots_text.setText(str(max_spike_plots))
        self.max_spike_plots_text.textChanged.connect(lambda: maxSpikesChange(self, 'main'))

        max_spikes_layout = QtWidgets.QHBoxLayout()
        max_spikes_layout.addWidget(max_spike_label)
        max_spikes_layout.addWidget(self.max_spike_plots_text)

        # ------- move invalid -----------

        move_to_channel_label = QtWidgets.QLabel("Move to Cell:")
        self.move_to_channel = QtWidgets.QLineEdit()
        self.move_to_channel.setToolTip("This is the channel that you will move selected cells too.")
        self.move_to_channel.setText(default_move_channel)
        self.move_to_channel.textChanged.connect(lambda: moveToChannel(self, 'main'))

        move_to_channel_layout = QtWidgets.QHBoxLayout()
        move_to_channel_layout.addWidget(move_to_channel_label)
        move_to_channel_layout.addWidget(self.move_to_channel)

        # -------- spike parameter options -------------------------------------

        feature_options = ["None"] + list(feature_name_map.keys())

        spike_parameter_layout = QtWidgets.QHBoxLayout()
        x_axis_label = QtWidgets.QLabel("X-Axis")
        self.x_axis_cb = QtWidgets.QComboBox()
        for option in feature_options:
            self.x_axis_cb.addItem(option)
        self.x_axis_cb.setCurrentIndex(self.x_axis_cb.findText(defaultXAxis))

        self.x_axis_cb.setToolTip("Choose a feature to plot on the X-Axis")

        x_axis_layout = QtWidgets.QHBoxLayout()
        x_axis_layout.addWidget(x_axis_label)
        x_axis_layout.addWidget(self.x_axis_cb)

        y_axis_label = QtWidgets.QLabel("Y-Axis")
        self.y_axis_cb = QtWidgets.QComboBox()
        self.y_axis_cb.setToolTip("Choose a feature to plot on the Y-Axis")
        for option in feature_options:
            self.y_axis_cb.addItem(option)
        self.y_axis_cb.setCurrentIndex(self.y_axis_cb.findText(defaultYAxis))

        y_axis_layout = QtWidgets.QHBoxLayout()
        y_axis_layout.addWidget(y_axis_label)
        y_axis_layout.addWidget(self.y_axis_cb)

        z_axis_label = QtWidgets.QLabel("Z-Axis")
        self.z_axis_cb = QtWidgets.QComboBox()
        self.z_axis_cb.setToolTip("Choose a feature to plot on the Z-Axis")
        for option in feature_options:
            self.z_axis_cb.addItem(option)
        self.z_axis_cb.setCurrentIndex(self.z_axis_cb.findText(defaultZAxis))

        z_axis_layout = QtWidgets.QHBoxLayout()
        z_axis_layout.addWidget(z_axis_label)
        z_axis_layout.addWidget(self.z_axis_cb)

        axis_layout = QtWidgets.QHBoxLayout()
        axis_layout.addLayout(x_axis_layout)
        axis_layout.addLayout(y_axis_layout)
        axis_layout.addLayout(z_axis_layout)

        tetrode_layout = QtWidgets.QHBoxLayout()
        tetrode_label = QtWidgets.QLabel("Tetrode:")
        self.tetrode_cb = QtWidgets.QComboBox()
        self.tetrode_cb.currentIndexChanged.connect(self.tetrode_changed)
        tetrode_layout.addWidget(tetrode_label)
        tetrode_layout.addWidget(self.tetrode_cb)

        spike_parameter_widgets = [tetrode_layout, axis_layout, move_to_channel_layout, max_spikes_layout]

        spike_parameter_layout.addStretch(1)
        for i, widget in enumerate(spike_parameter_widgets):
            if 'Layout' in widget.__str__():
                spike_parameter_layout.addLayout(widget)
                spike_parameter_layout.addStretch(1)
            else:
                spike_parameter_layout.addWidget(widget, 0, QtCore.Qt.AlignCenter)
                spike_parameter_layout.addStretch(1)

        if openGL:
            self.feature_win = pg.GraphicsWindow()
            feature_win_layout = QtWidgets.QGridLayout()
            self.feature_win.setLayout(feature_win_layout)
            self.glViewWidget = gl.GLViewWidget()
            self.glViewWidget.setBackgroundColor('k')

            feature_win_layout.addWidget(self.glViewWidget)
        else:
            # self.feature_win = PltWidget(self)
            self.feature_win = MatplotlibWidget()
            self.feature_win.toolbar.hide()  # hide the toolbar

        self.unit_win = pg.GraphicsWindow()

        plot_layout = QtWidgets.QHBoxLayout()
        for _object in [self.feature_win, self.unit_win]:
            plot_layout.addWidget(_object)

        # --------- Create the Buttons at the bottom of the Main Window ------------- #

        button_layout = QtWidgets.QHBoxLayout()

        self.plot_btn = QtWidgets.QPushButton("Plot")
        self.plot_btn.clicked.connect(self.plotFunc)

        # self.reload_cut_btn = QtWidgets.QPushButton("Reload Cut/Clu")
        # self.reload_cut_btn.clicked.connect(self.reload_cut)

        self.save_btn = QtWidgets.QPushButton("Save Cut")
        self.save_btn.clicked.connect(self.save_function)
        self.save_btn.setToolTip('Save to the current cut file')

        self.undo_btn = QtWidgets.QPushButton("Undo")
        self.undo_btn.clicked.connect(lambda: undo_function(self))
        self.undo_btn.setToolTip('Click to undo previous action!')

        self.quit_btn = QtWidgets.QPushButton("Quit")
        self.quit_btn.clicked.connect(self.close_app)
        self.quit_btn.setShortcut("Ctrl+Q")  # creates shortcut for the quit button
        self.quit_btn.setToolTip('Click to quit gebaSpike!')

        button_order = [self.plot_btn,
                        self.save_btn,
                        self.undo_btn,
                        # self.reload_cut_btn,
                        self.quit_btn]

        for btn in button_order:
            button_layout.addWidget(btn)

        main_window_layout = QtWidgets.QVBoxLayout()

        layout_order = [filename_grid_layout, spike_parameter_layout, plot_layout, button_layout]
        add_Stretch = [False, False, False, False]
        # ---------------- add all the layouts and widgets to the Main Window's layout ------------ #

        # main_window_layout.addStretch(1)  # adds the widgets/layouts according to the order
        for widget, addStretch in zip(layout_order, add_Stretch):
            if 'Layout' in widget.__str__():
                main_window_layout.addLayout(widget)
                if addStretch:
                    main_window_layout.addStretch(1)
            else:
                main_window_layout.addWidget(widget, 0, QtCore.Qt.AlignCenter)
                if addStretch:
                    main_window_layout.addStretch(1)

        self.setLayout(main_window_layout)  # defining the layout of the Main Window

        self.show()  # shows the window

    def addPopup(self, cell):
        self.PopUpCutWindow[cell] = PopUpCutWindow(self)

    def save_function(self):

        if self.cut_filename.text() == default_filename:
            return

        save_filename = os.path.realpath(self.cut_filename.text())

        if os.path.exists(save_filename):
            self.choice = None
            self.LogError.signal.emit('OverwriteCut!%s' % save_filename)
            while self.choice is None:
                time.sleep(0.1)

            if self.choice != QtWidgets.QMessageBox.Yes:
                return

        if len(self.tetrode_data) == 0:
            return

        # organize the cut data
        n_spikes_expected = self.tetrode_data.shape[1]
        n_spikes = len(np.asarray([item for sublist in self.cell_indices.values() for item in sublist]))

        # check that with the manipulation of the spikes, that we still have the correct number of spikes
        if n_spikes != n_spikes_expected:
            self.choice = None
            self.LogError.signal.emit('cutSizeError')
            while self.choice is None:
                time.sleep(0.1)
            return

        # we will check if we are missing some of the spikes somehow. If we kept track of them, then the indices from
        # the spikes, when sorted, should produce an array from 0 -> N-1 spikes.
        if not np.array_equal(np.sort(np.asarray([item for sublist in self.cell_indices.values() for item in sublist])),
                          np.arange(len(self.cut_data_original))):
            self.choice = None
            self.LogError.signal.emit('cutIndexError')
            while self.choice is None:
                time.sleep(0.1)
            return

        cut_values = np.zeros(n_spikes)
        for cell, cell_indices in self.cell_indices.items():
            cut_values[cell_indices] = cell

        if '.clu.' in save_filename:
            # save the .clu filename
            write_clu(save_filename, cut_values)
            self.choice = None
            self.LogError.signal.emit('saveCompleteClu')
            while self.choice is None:
                time.sleep(0.1)
            return
        else:
            # save the cut filename
            write_cut(save_filename, cut_values)
            self.choice = None
            self.LogError.signal.emit('saveComplete')
            while self.choice is None:
                time.sleep(0.1)
            return

    def close_app(self):
        """This method will prompt the user, asking if they would like to quit or not"""
        # pop up window that asks if you really want to exit the app ------------------------------------------------

        choice = QtWidgets.QMessageBox.question(self, "Quitting ",
                                            "Do you really want to exit?",
                                            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if choice == QtWidgets.QMessageBox.Yes:
            sys.exit()  # tells the app to quit
        else:
            pass

    def raiseError(self, error):
        if 'TetrodeExistError' in error:
            filename = error.split('!')[1]
            self.choice = QtWidgets.QMessageBox.question(self, "Tetrode Filename Does Not Exist!",
                                                         "The following tetrode filename does not exist: \n%s\nPlease" %
                                                         filename +
                                                         " ensure that the file exists before attempting to plot the"
                                                         " data",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'InvalidSession' in error:
            session = error.split('!')[1]
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Session Filename!",
                                                         "The following session filename is invalid: \n%s\nPlease" %
                                                         session +
                                                         " ensure that the appropriate files exist for this session.",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'cutIndexError' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Cut Index Error!",
                                                         "The cut output is missing some of the spikes!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'cutSizeError' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Cut Size Error!",
                                                         "Trying to save an inappropriate number of spikes, cannot " +
                                                         "save cut file!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'InvalidCut' in error:
            session = error.split('!')[1]
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Cut/Clu Filename!",
                                                         "The following Cut filename is invalid: \n%s\nPlease" %
                                                         session +
                                                         " ensure that this .cut/clu filename belongs to this session!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'ChooseSession' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Choose Session Filename!",
                                                         "You have not chosen a session file yet! Please choose a"
                                                         " session filename before proceeding!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'InvalidMoveChannel' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Move to Cell Value!",
                                                         "The value you have chosen for the 'Move to Cell' value is "
                                                         "invalid, please choose a valid value before continuing!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'SameChannelInvalid' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Same Channel Error!",
                                                         "The value you have chosen for the 'Move to Cell' value is "
                                                         "the same as the cell you are cutting from! If you would like "
                                                         "to move these selected spikes to a different channel, please "
                                                         "choose another channel!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'ActionsMade' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Are you sure...?",
                                                         "You have performed some actions that will be lost when you"
                                                         " reload this cut file. Are you sure you want to continue?",
                                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        elif 'OverwriteCut' in error:
            cut_file = os.path.realpath(error.split('!')[1])
            self.choice = QtWidgets.QMessageBox.question(self, "Cut Filename Exists",
                                                         "The following cut filename exists:\n\n%s\n\n Are you sure you" 
                                                         " want to overwrite this file?" % cut_file,
                                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        elif 'saveComplete' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Save Complete",
                                                         "Your '.cut' file has been saved successfully!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'saveCompleteClu' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Save Complete",
                                                         "Your '.clu' file has been saved successfully!",
                                                         QtWidgets.QMessageBox.Ok)

        elif 'invalidMaxSpikes' in error:
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Max Spikes",
                                                         "The number chosen for Max Spikes is invalid!",
                                                         QtWidgets.QMessageBox.Ok)

    def get_settings(self):
        settings = {}
        if os.path.exists(self.settings_filename):
            with open(self.settings_filename, 'r') as f:
               settings = json.load(f)
        return settings

    def overwrite_settings(self):
        # overwrite settings file
        with open(self.settings_filename, 'w') as f:
            json.dump(self.settings, f, sort_keys=True, indent=4)

    def reset_parameters(self):
        self.feature_data = None
        self.tetrode_data = None
        self.cut_data = None
        self.cut_data_original = None
        self.n_channels = None
        self.spike_times = None

        self.max_spike_plots = None

        self.cell_indices = {}
        self.plot_lines = {}
        self.avg_plot_lines = {}
        self.vb = {}
        self.active_ROI = []
        self.unit_data = {}
        self.cell_subsample_i = {}
        self.unit_positions = {}
        self.original_cell_count = {}

        self.PopUpCutWindow = {}

        self.latest_actions = {}

        self.actions_made = False

        self.last_drag_index = None

        self.drag_active = False

        self.unit_drag_lines = {}

        self.samples_per_spike = None
        self.spike_colors = None

        self.unit_plots = {}
        self.unit_rows = 0
        self.unit_cols = 0

        self.reset_plots()

    def reset_plots(self):
        if hasattr(self, 'scatterItem'):
            if self.scatterItem is not None:
                self.glViewWidget.removeItem(self.scatterItem)
                self.scatterItem = None
        self.unit_win.clear()

    def choose_cut_filename(self):
        if 'file_directory' not in self.settings.keys():
            current_filename, filename_filter = QtWidgets.QFileDialog.getOpenFileName(
                self, caption="Select a '.Set' file!", directory='', filter='Set Files (*.set)')

        else:
            if os.path.exists(self.settings['file_directory']):
                current_filename, filename_filter = QtWidgets.QFileDialog.getOpenFileName(
                    self, caption="Select a '.Cut/.Clu' file!", directory=self.settings['file_directory'],
                    filter='Cut Files (*.cut, *.clu*)')
            else:
                current_filename, filename_filter = QtWidgets.QFileDialog.getOpenFileName(
                    self, caption="Select a '.Cut/.Clu' file!", directory='', filter='Cut Files (*.cut, *.clu*)')

        # if no file chosen, skip
        if current_filename == '':
            return

        chosen_directory = os.path.dirname(current_filename)

        self.settings['file_directory'] = chosen_directory

        self.overwrite_settings()

        cut_valid, error_raised = validate_cut(self, self.filename.text(), current_filename)

        if cut_valid:
            # replace the current .cut field in the choose .cut field with chosen filename
            self.cut_filename.setText(current_filename)

        else:
            if not error_raised:
                self.choice = None
                self.LogError.signal.emit('InvalidCut!%s' % current_filename)
                while self.choice is None:
                    time.sleep(0.1)

    def plotFunc(self):
        if self.actions_made is True:
            self.choice = None
            self.LogError.signal.emit('ActionsMade')
            while self.choice is None:
                time.sleep(0.1)

            if self.choice == QtWidgets.QMessageBox.Yes:
                self.reset_plots()
                self.reset_parameters()
                plot_session(self)
        else:
            plot_session(self)

    def choose_filename(self):
        """
        This method will allow you to choose a filename to analyze.
        :return:
        """

        if 'file_directory' not in self.settings.keys():
            current_filename, filename_filter = QtWidgets.QFileDialog.getOpenFileName(
                self, caption="Select a '.Set' file!", directory='', filter='Set Files (*.set)')

        else:
            if os.path.exists(self.settings['file_directory']):
                current_filename, filename_filter = QtWidgets.QFileDialog.getOpenFileName(
                    self, caption="Select a '.Set' file!", directory=self.settings['file_directory'],
                    filter='Set Files (*.set)')
            else:
                current_filename, filename_filter = QtWidgets.QFileDialog.getOpenFileName(
                    self, caption="Select a '.Set' file!", directory='', filter='Set Files (*.set)')

        # if no file chosen, skip
        if current_filename == '':
            return

        chosen_directory = os.path.dirname(current_filename)

        self.settings['file_directory'] = chosen_directory

        self.overwrite_settings()

        session_valid, error_raised = validate_session(self, current_filename)

        if session_valid:
            # replace the current .set field in the choose .set field with chosen filename
            self.filename.setText(os.path.realpath(current_filename))

            self.reset_parameters()

        else:
            if not error_raised:
                self.choice = None
                self.LogError.signal.emit('InvalidSession!%s' % current_filename)
                while self.choice is None:
                    time.sleep(0.1)

    def set_cut_filename(self):
        filename = self.filename.text()

        try:
            tetrode = int(self.tetrode_cb.currentText())
        except ValueError:
            return

        cut_filename = '%s_%d.cut' % (os.path.splitext(filename)[0], tetrode)
        self.cut_filename.setText(cut_filename)

    def tetrode_changed(self):

        # we will update the cut_filename
        self.set_cut_filename()

    def filename_changed(self):
        """
        This method will run when the filename LineEdit has been changed.

        It will essentially find the active tetrodes and populate the drop-down menu.
        """

        filename = self.filename.text()
        if os.path.exists(filename):
            self.tetrode_cb.clear()

            tetrode_list = find_tetrodes(self.filename.text())

            for file in tetrode_list:
                tetrode = os.path.splitext(file)[-1][1:]
                self.tetrode_cb.addItem(tetrode)

        self.set_cut_filename()


def launch_gui():
    """
    This function will launch the gebaSpike GUI
    :return:
    """

    app = QtWidgets.QApplication(sys.argv)

    main_window = MainWindow()  # Creating the Main Window
    main_window.raise_()  # raises the Main Window

    sys.exit(app.exec_())  # prevents the window from immediately exiting out


if __name__ == '__main__':
    launch_gui()
